# Copyright (C) 2021 Hyunwoong Ko <kevin.ko@tunib.ai> and Sang-Kil Park <skpark1224@hyundai.com>
# All rights reserved.

from kss._modules.morphemes.split_morphemes import split_morphemes
from kss._modules.sentences.split_sentences import split_sentences
from kss._modules.summarization.summarize_sentences import summarize_sentences

__ALL__ = [split_sentences, split_morphemes, summarize_sentences]
__version__ = "4.5.4"



# Copyright (C) 2021 Hyunwoong Ko <kevin.ko@tunib.ai> and Sang-Kil Park <skpark1224@hyundai.com>
# All rights reserved.

from abc import ABC
from functools import lru_cache
from typing import Tuple, List, Any

from kss._modules.morphemes.utils import _get_mecab, _get_pecab, _preserve_space
from kss._utils.const import spaces


class Analyzer(ABC):
    _analyzer, _backend = None, None

    def pos(self, text: str, drop_space: bool) -> Any:
        raise NotImplementedError

    @staticmethod
    def _drop_space(tokens: List[Tuple[str, str]]) -> List[Tuple[str, str]]:
        return [token for token in tokens if token[0] not in spaces]


class MecabAnalyzer(Analyzer):
    # `_analyzer` object must be class variable because of multiprocessing
    _analyzer, _backend = _get_mecab()

    @lru_cache(maxsize=500)
    def pos(self, text: str, drop_space: bool) -> List[Tuple[str, str]]:
        """
        Get pos information.

        Args:
            text (str): input text
            drop_space (bool): drop all spaces or not.

        Returns:
            List[Tuple[str, str]]: output of analysis.
        """
        output = self._analyzer.pos(text)
        output = _preserve_space(text, output, spaces=" \n\r\t\v")

        if drop_space:
            output = self._drop_space(output)

        return output


class PecabAnalyzer(Analyzer):
    # `_analyzer` object must be class variable because of multiprocessing
    _analyzer, _backend = _get_pecab()

    @lru_cache(maxsize=500)
    def pos(self, text: str, drop_space: bool) -> List[Tuple[str, str]]:
        """
        Get pos information.

        Args:
            text (str): input text
            drop_space (bool): drop all spaces or not.

        Returns:
            List[Tuple[str, str]]: output of analysis.
        """
        output = self._analyzer.pos(text)
        output = _preserve_space(text, output, spaces=" \n\r\t\v\f")

        if drop_space:
            output = self._drop_space(output)

        return output


class CharacterAnalyzer(Analyzer):
    _analyzer, _backend = None, "character"

    @lru_cache(maxsize=500)
    def pos(self, text: str, drop_space: bool) -> List[Tuple[str, str]]:
        """
        Get pos information.

        Args:
            text (str): input text
            drop_space (bool): drop all spaces or not.

        Returns:
            List[Tuple[str, str]]: output of analysis.
        """
        output = [(char, "-") for char in text]

        if drop_space:
            output = self._drop_space(output)

        return output



# Copyright (C) 2021 Hyunwoong Ko <kevin.ko@tunib.ai> and Sang-Kil Park <skpark1224@hyundai.com>
# All rights reserved.

from functools import lru_cache
from itertools import chain
from typing import List, Dict, Tuple, Iterable, Optional

from kss._elements.subclasses import Syllable
from kss._utils.const import (
    double_quotes,
    double_quotes_open_to_close,
    double_quotes_close_to_open,
    single_quotes,
    single_quotes_open_to_close,
    single_quotes_close_to_open,
    single_quotes_wo_direction,
    double_quotes_wo_direction,
)


class EmbracingProcessor:
    def __init__(self):
        self.single_stack, self.double_stack = [], []
        self.single_pop, self.double_pop = "'", '"'
        self.single_idx, self.double_idx = 0, 0
        self.single_sent_idx, self.double_sent_idx = 0, 0

    def empty(self) -> bool:
        """
        Check all stacks are empty or not.

        Returns:
            bool: True if all stacks are empty else False.
        """
        return self._empty([self.single_stack, self.double_stack], dim=2)

    def process(self, idx: int, sent_idx: int, syllable: Syllable):
        """
        Push or pop symbols to detect embraced sentences

        Args:
            idx (int): current syllable index
            sent_idx (int): current sentence index
            syllable (Syllable): current syllable object
        """
        if syllable.text in single_quotes_wo_direction:
            self.single_sent_idx = sent_idx
            self.single_idx = idx
            self.single_pop = self._pop_symbol(
                syllable=syllable,
                stack=self.single_stack,
                open_to_close=single_quotes_open_to_close,
                close_to_open=single_quotes_close_to_open,
            )
        elif syllable.text in double_quotes_wo_direction:
            self.double_sent_idx = sent_idx
            self.double_idx = idx
            self.double_pop = self._pop_symbol(
                syllable=syllable,
                stack=self.double_stack,
                open_to_close=double_quotes_open_to_close,
                close_to_open=double_quotes_close_to_open,
            )

    def update_index(self, idx: int, sent_idx: int, syllable: Syllable):
        """
        Update indices of syllables and sentences

        Args:
            idx (int): current syllable index
            sent_idx (int): current sentence index
            syllable (Syllable): current syllable object
        """

        if syllable.text in single_quotes:
            self.single_idx = idx
            self.single_sent_idx = sent_idx

        elif syllable.text in double_quotes:
            self.double_idx = idx
            self.double_sent_idx = sent_idx

    def realign(
        self,
        input_sentences: List[Syllable],
        output_sentences: List[List[Syllable]],
        func: "function",
    ) -> List[List[Syllable]]:
        """
        Realign wrongly split sentences because of all symbols.

        Args:
            input_sentences (List[Syllable]): input sentences
            output_sentences (List[List[Syllables]): split sentences from `_split_sentences`
            func (function): split function

        Returns:
            List[List[Syllable]]: corrected split sentences.
        """
        if len(self.single_stack) != 0:
            return self._realign_sentences(
                input_sentences=input_sentences,
                output_sentences=output_sentences,
                idx=self.single_idx,
                sent_idx=self.single_sent_idx,
                func=func,
            )

        if len(self.double_stack) != 0:
            return self._realign_sentences(
                input_sentences=input_sentences,
                output_sentences=output_sentences,
                idx=self.double_idx,
                sent_idx=self.double_sent_idx,
                func=func,
            )

        return output_sentences

    def _pop_symbol(
        self,
        syllable: Syllable,
        stack: List[str],
        open_to_close: Dict[str, str],
        close_to_open: Dict[str, str],
    ) -> str:
        """
        Pop symbols from given stack if the symbols are contained in given dictionaries.

        Args:
            syllable (Syllable): syllable object
            stack (List[str]): symbol stack
            open_to_close (Dict[str, str]): open to close dict
            close_to_open (Dict[str, str]): close to open dict

        Returns:
            str: popped symbol
        """
        if syllable.text in open_to_close.keys():
            pop = self._push_pop_symbol(
                stack=stack,
                symbol=open_to_close[syllable.text],
                current_char=syllable.text,
            )
        else:
            pop = self._push_pop_symbol(
                stack=stack,
                symbol=close_to_open[syllable.text],
                current_char=syllable.text,
            )
        return pop

    def _realign_sentences(
        self,
        input_sentences: List[Syllable],
        output_sentences: List[List[Syllable]],
        idx: int,
        sent_idx: int,
        func: "function",
    ) -> List[List[Syllable]]:
        """
        Realign wrongly split sentences because of specific symbol.

        Args:
            input_sentences (List[Syllable]): input sentences
            output_sentences (List[List[Syllables]): split sentences from `_split_sentences`
            idx (int): current syllable index
            sent_idx (int): current sentence index
            func (function): split function

        Returns:
            List[List[Syllable]]: corrected split sentences.
        """
        return output_sentences[:sent_idx] + self._realign_sub_sentences(
            output_sentences=tuple(chain(*output_sentences[sent_idx:])),
            syllable=input_sentences[idx],
            idx_in_sent=self.get_idx_in_sent(output_sentences, idx, sent_idx),
            func=func,
        )

    @lru_cache(maxsize=500)
    def _realign_sub_sentences(
        self,
        output_sentences: Tuple,
        syllable: Syllable,
        idx_in_sent: int,
        func: "function",
    ):
        """
        Realign wrongly split sub-sentences because of specific symbol.

        Args:
            output_sentences (Tuple): tuple of syllables
            syllable (Syllable): problematic syllable
            idx_in_sent (int): syllable index in sentence
            func (function): split function

        Returns:
            List[List[Syllable]]: corrected split sub-sentences.
        """
        before_quote = func(output_sentences[:idx_in_sent])
        before_last = before_quote[-1] if len(before_quote) > 0 else []
        before_quote = [] if len(before_quote) == 1 else before_quote[:-1]

        after_quote = func(output_sentences[idx_in_sent + 1 :])
        after_first = after_quote[0] if len(after_quote) > 0 else []
        after_quote = [] if len(after_first) == 1 else after_quote[1:]

        middle_quote = [before_last + [syllable] + after_first]
        return before_quote + middle_quote + after_quote

    @staticmethod
    def get_idx_in_sent(
        output_sentences: List[List[Syllable]],
        idx: int,
        sent_idx: int,
    ) -> int:
        """
        Get syllable index in the sentence.

        Args:
            output_sentences (List[List[Syllables]): split sentences from `_split_sentences`
            idx (int): current syllable index
            sent_idx (int): current sentence index

        Returns:
            int: syllable index in sentence
        """
        return idx - (
            0 if sent_idx == 0 else sum([len(o) for o in output_sentences[:sent_idx]])
        )

    @staticmethod
    def _top(stack: List[str], symbol: str) -> bool:
        """
        Is the symbol was top of the stack or not.

        Args:
            stack (List[str]): stack of symbols
            symbol (symbol): symbol string

        Returns:
            bool: whether the symbol was top of the stack or not.
        """
        return stack[len(stack) - 1] == symbol

    def _empty(self, obj: Iterable, dim: int = 1) -> bool:
        """
        Check the length of object is 0.

        Args:
            obj (Iterable): iterable object
            dim (int): object dimension

        Returns:
            bool: whether the length of object is 0 or not.
        """
        assert dim in [1, 2], "only 1 or 2 dimension iterable is supported."

        if dim == 1:
            return len(obj) == 0
        else:
            return all([self._empty(o) for o in obj])

    def _push_pop_symbol(
        self,
        stack: List[str],
        symbol: str,
        current_char: str,
    ) -> Optional[str]:
        """
        Push or pop symbol in the list.

        Args:
            stack (List[str]): stack of symbols
            symbol (str): symbol string
            current_char (str): current character of the syllable

        Returns:
            Optional[str]: popped symbol string if the symbol was top, else returns None.
        """
        if self._empty(stack):
            stack.append(symbol)

        else:
            if self._top(stack, current_char):
                return stack.pop()
            else:
                stack.append(symbol)

        return current_char



# Copyright (C) 2021 Hyunwoong Ko <kevin.ko@tunib.ai> and Sang-Kil Park <skpark1224@hyundai.com>
# All rights reserved.
import re
from functools import lru_cache
from typing import List

from kss._utils.const import (
    alphabet_with_quotes,
    url_pattern,
    email_pattern,
    backup_normal,
)


class SentenceProcessor:
    _all_s_exclude = ("QTO",)
    _all_s_poses = ("SP", "SF", "SY", "SE", "SSC", "QTC", "QTN", "EMOJI", "JAMO")
    _all_s_poses_wo_qtn = ("SP", "SF", "SY", "SE", "SSC", "QTC", "EMOJI", "JAMO")

    _heavy_backup = {}
    _heavy_backup.update(
        {
            k: {_v: str(abs(hash(_v))) for _v in v}
            for k, v in alphabet_with_quotes.items()
        }
    )
    _normal_backup = {k: str(abs(hash(k))) for k in sorted(backup_normal)}

    def __init__(self, ignores: List[str] = None):
        self.ignores = ignores
        if self.ignores is not None:
            self._normal_backup.update({k: str(abs(hash(k))) for k in self.ignores})

    @staticmethod
    def _replace(text: str, purpose_dict: dict, restore: bool = False):
        for k, v in purpose_dict.items():
            if restore is True:
                source, target = v, k
            else:
                source, target = k, v

            if source in text:
                text = text.replace(source, target)

        return text

    def _add_url_or_email(self, text):
        _url_or_email = {
            k: str(abs(hash(k)))
            for k in re.findall(url_pattern, text) + re.findall(email_pattern, text)
        }
        self._normal_backup.update(_url_or_email)

    @lru_cache(100)
    def backup(self, inputs: str):
        self._add_url_or_email(inputs)
        inputs = self._replace(inputs, self._normal_backup)

        for source, purpose_dict in self._heavy_backup.items():
            if source in inputs:
                inputs = self._replace(inputs, purpose_dict, restore=False)

        return inputs

    @lru_cache(100)
    def restore(self, outputs: str, inputs: str):
        outputs = self._replace(outputs, self._normal_backup, restore=True)

        for source, purpose_dict in self._heavy_backup.items():
            if source in inputs:
                outputs = self._replace(outputs, purpose_dict, restore=True)

        return outputs



# Copyright (C) 2021 Hyunwoong Ko <kevin.ko@tunib.ai> and Sang-Kil Park <skpark1224@hyundai.com>
# All rights reserved.

from dataclasses import dataclass
from typing import Optional, List, Tuple


@dataclass
class Empty(object):
    _next = None
    _prev = None

    def __init__(self, text="", pos="", idx=-1):
        self.text = text
        self.pos = pos
        self.idx = idx

    def __str__(self):
        return f"{self.__class__.__qualname__}()"

    def __repr__(self):
        return self.__str__()

    @property
    def next(self):
        return self

    @property
    def prev(self):
        return self

    @next.setter
    def next(self, _next):
        pass

    @prev.setter
    def prev(self, _prev):
        pass

    def next_skip(self, *poses, exclude=None):
        return self

    def prev_skip(self, *poses, exclude=None):
        return self

    def prev_skip_from_current(self, *poses, exclude=None):
        return self

    def next_skip_from_current(self, *poses, exclude=None):
        return self

    def check_pos(self, *poses, exclude: Optional[List[str]] = None) -> bool:
        """
        Check pos of given syllable.

        Args:
            poses (Tuple[str]): poses for check
            exclude (Optional[List[str]]): excluded poses

        Returns:
            bool: whether pos of the syllable is contained in poses or not.
        """
        return False

    def check_text(self, *texts, exclude: Optional[List[str]] = None) -> bool:
        """
        Check text of given syllable.

        Args:
            texts (Tuple[str]): texts for check
            exclude (Optional[List[str]]): excluded poses

        Returns:
            bool: whether text of the syllable is contained in pos_list or not.
        """
        return False

    def check_pos_and_text(
        self,
        poses: Tuple,
        texts: Tuple,
        exclude_poses: Optional[Tuple] = None,
        exclude_texts: Optional[Tuple] = None,
    ):
        """
        Check pos and text at the same time

        Args:
            poses: poses for check
            texts (Tuple): texts for check
            exclude_poses (Optional[Tuple]): excluded poses
            exclude_texts (Optional[Tuple]): excluded texts

        Returns:
            bool: whether pos and text of the syllable are contained in input poses and texts.
        """

    def check_texts(self, text: str) -> bool:
        """
        Check texts of current and next syllables

        Args:
            text (str): texts for check

        Returns:
            bool: whether text of the current and next syllables re contained in input text or not.
        """
        return False



# Copyright (C) 2021 Hyunwoong Ko <kevin.ko@tunib.ai> and Sang-Kil Park <skpark1224@hyundai.com>
# All rights reserved.

import emoji
import regex

_emojis = {}
_specials = ["♡", "♥"]
_unicodes = [
    "‍",  # Zero width joiner
    "︀",  # Variation Selector-1
    "︁",  # Variation Selector-2
    "︂",  # Variation Selector-3
    "︃",  # Variation Selector-4
    "︄",  # Variation Selector-5
    "︅",  # Variation Selector-6
    "︆",  # Variation Selector-7
    "︇",  # Variation Selector-8
    "︈",  # Variation Selector-9
    "︉",  # Variation Selector-10
    "︊",  # Variation Selector-11
    "︋",  # Variation Selector-12
    "︌",  # Variation Selector-13
    "︍",  # Variation Selector-14
    "︎",  # Variation Selector-15
    "️",  # Variation Selector-16
]

try:
    for lang in ["pt", "it", "es", "en"]:
        _emojis.update(emoji.unicode_codes.UNICODE_EMOJI[lang])
        _emojis.update({k: "" for k in _specials})
        _emojis.update({k: "" for k in _unicodes})
except Exception as e:
    raise ImportError("Kss requires `emoji==1.2.0`. please install that version.")


def get_emoji(text):
    emoji_list = []
    flags = regex.findall("[\U0001F1E6-\U0001F1FF]", text)

    for grapheme in regex.findall(r"\X", text):
        if any(char in _emojis for char in grapheme):
            emoji_list.append(grapheme)

    return emoji_list + flags



# This code was copied from Text Rank KR [https://github.com/theeluwin/textrankr]
# And modified by Hyunwoong Ko [https://github.com/hyunwoongko]

from functools import partial
from typing import List, Dict, Union, Tuple

from kss._modules.summarization.sentence import Sentence
from kss._modules.summarization.utils import (
    _parse_text_into_sentences,
    _build_sentence_graph,
)
from kss._utils.multiprocessing import _run_job
from kss._utils.sanity_checks import (
    _check_text,
    _check_num_workers,
    _check_analyzer_backend,
    _check_type,
    _check_value,
)


def summarize_sentences(
    text: Union[str, List[str], Tuple[str]],
    backend: str = "auto",
    num_workers: Union[str, int] = "auto",
    max_sentences: int = 3,
    tolerance: Union[float] = 0.05,
    strip: bool = True,
    ignores: List[str] = None,
) -> Union[List[str], List[List[str]]]:
    """
    Summarizes the given text, using TextRank algorithm.

    Args:
        text (Union[str, List[str], Tuple[str]]): single text or list/tuple of texts
        backend (str): morpheme analyzer backend. 'mecab', 'pecab' are supported.
        num_workers (Union[int, str])): the number of multiprocessing workers
        max_sentences (int): the max number of sentences in a summarization result.
        tolerance (float): a threshold for omitting edge weights.
        strip (bool): strip all sentences or not
        ignores (List[str]): list of strings to ignore

    Returns:
        Union[List[str], List[List[str]]]: outputs of text summarization
    """
    text, finish = _check_text(text)

    if finish:
        return text

    _check_analyzer_backend(backend)
    num_workers = _check_num_workers(text, num_workers)
    max_sentences = _check_type(max_sentences, "max_sentences", int)
    max_sentences = _check_value(
        max_sentences, "max_sentences", lambda x: x > 0, "integer value in 1~N"
    )
    tolerance = _check_type(tolerance, "tolerance", float)

    return _run_job(
        func=partial(
            _summarize_sentences,
            backend=backend,
            max_sentences=max_sentences,
            tolerance=tolerance,
            strip=strip,
            ignores=ignores,
        ),
        inputs=text,
        num_workers=num_workers,
    )


def _summarize_sentences(
    text: str,
    backend: str = "auto",
    max_sentences: int = 3,
    tolerance: float = 0.05,
    strip: bool = True,
    ignores: List[str] = None,
) -> List[str]:
    """
    Summarizes the given text, using TextRank algorithm.

    Args:
        text (str): a single text to be summarized.
        backend (str): morpheme analyzer backend. 'mecab', 'pecab' are supported.
        max_sentences (int): the max number of sentences in a summarization result.
        tolerance (float): a threshold for omitting edge weights.
        strip (bool): strip all sentences or not
        ignores (List[str]): list of strings to ignore

    Returns:
        List[str]: summarized sentences
    """
    from networkx import pagerank

    # parse text
    sentences: List[Sentence] = _parse_text_into_sentences(text, backend, strip, ignores)

    # build graph
    graph = _build_sentence_graph(sentences, tolerance=tolerance)

    # run pagerank
    pageranks: Dict[Sentence, float] = pagerank(graph, weight="weight")

    # get top-k sentences
    sentences = sorted(pageranks, key=pageranks.get, reverse=True)
    sentences = sentences[:max_sentences]
    sentences = sorted(sentences, key=lambda sentence: sentence.index)

    # output summaries
    return [sentence.text for sentence in sentences]



# This code was copied from Text Rank KR [https://github.com/theeluwin/textrankr]
# And modified by Hyunwoong Ko [https://github.com/hyunwoongko]

from collections import Counter


class Sentence:
    """
    Notes:
        The purpose of this class is as follows:
            1. In order to use the 'pagerank' function in the networkx library, you need a hashable object.
            2. Summaries should keep the sentence order from its original text to improve the verbosity.

        Note that the 'bow' stands for 'bag-of-words'.
    """

    def __init__(self, index: int, text: str, bow: Counter) -> None:
        self.index: int = index
        self.text: str = text
        self.bow: Counter = bow

    def __str__(self) -> str:
        return self.text

    def __hash__(self) -> int:
        return self.index



# Copyright (C) 2021 Hyunwoong Ko <kevin.ko@tunib.ai> and Sang-Kil Park <skpark1224@hyundai.com>
# All rights reserved.

"""
Sets of characters, contain method in set is much faster than list
"""
import re

# 0 ~ 9
numbers = set([str(_) for _ in range(0, 10)])

# a-z
lower_alphabets = set([chr(_) for _ in range(ord("a"), ord("z") + 1)])

# A-Z
upper_alphabets = set([_.upper() for _ in lower_alphabets])

# a-z|A-Z
alphabets = set()
alphabets.update(lower_alphabets)
alphabets.update(upper_alphabets)

# ㄱ-ㅎ
jaum = set([chr(_) for _ in range(ord("ㄱ"), ord("ㅎ") + 1)])

# ㅏ-ㅡ
moum = set([chr(_) for _ in range(ord("ㅏ"), ord("ㅡ") + 1)])

# ㄱ-ㅎ|ㅏ-ㅡ
jamo = set()
jamo.update(jaum)
jamo.update(moum)

# brackets | `<`, `>` 제외 (수학기호, 화살표로 쓰임)
brackets = {
    ")",
    "）",
    "〉",
    ">",
    "》",
    "]",
    "］",
    "〕",
    "】",
    "}",
    "｝",
    "』",
    "」",
    "(",
    "（",
    "〈",
    "<",
    "《",
    "[",
    "［",
    "〔",
    "【",
    "{",
    "｛",
    "「",
    "『",
}

# open -> close
bracket_open_to_close = {
    "(": ")",
    "（": "）",
    "〈": "〉",
    "<": ">",
    "《": "》",
    "[": "]",
    "［": "］",
    "〔": "〕",
    "【": "】",
    "{": "}",
    "｛": "｝",
    "「": "」",
    "『": "』",
}

# close -> open
bracket_close_to_open = {v: k for k, v in bracket_open_to_close.items()}

# doubles
double_quotes = {'"', "“", "”", "″"}

# doubles wo direction
double_quotes_wo_direction = {'"', "″"}

# singles
single_quotes = {"'", "‘", "’", "`"}

# singles wo direction
single_quotes_wo_direction = {"'", "`"}

# open to close
double_quotes_open_to_close = {"“": "”", '"': '"', "″": "″"}

# close to open
double_quotes_close_to_open = {v: k for k, v in double_quotes_open_to_close.items()}

# open to close
single_quotes_open_to_close = {"‘": "’", "'": "'", "`": "`"}

# close to open
single_quotes_close_to_open = {v: k for k, v in single_quotes_open_to_close.items()}

# open to close
quotes_open_to_close = dict()
quotes_open_to_close.update(single_quotes_open_to_close)
quotes_open_to_close.update(double_quotes_open_to_close)

# close to open
quotes_close_to_open = dict()
quotes_close_to_open.update(single_quotes_close_to_open)
quotes_close_to_open.update(double_quotes_close_to_open)

# quotes
quotes = set()
quotes.update(single_quotes)
quotes.update(double_quotes)

# quotes or brackets
quotes_or_brackets = set()
quotes_or_brackets.update(quotes)
quotes_or_brackets.update(brackets)

# opened brackets or quotes
opened_quotes_or_brackets = set(
    [
        _
        for _ in list(bracket_open_to_close.keys())
        + list(single_quotes_open_to_close.keys())
        + list(double_quotes_open_to_close.keys())
        if _ not in ['"', "'", "`"]
    ]
)

# closed brackets or quotes
closed_quotes_or_brackets = set(
    [
        _
        for _ in list(bracket_close_to_open.keys())
        + list(single_quotes_close_to_open.keys())
        + list(double_quotes_close_to_open.keys())
        if _ not in ['"', "'", "`"]
    ]
)

# open to close
quotes_or_brackets_open_to_close = dict()
quotes_or_brackets_open_to_close.update(quotes_open_to_close)
quotes_or_brackets_open_to_close.update(bracket_open_to_close)

# close to open
quotes_or_brackets_close_to_open = dict()
quotes_or_brackets_close_to_open.update(quotes_close_to_open)
quotes_or_brackets_close_to_open.update(bracket_close_to_open)

# special symbols
special_symbols_for_split = "㈜*∗¶§※○●◎◇◆□■△▲▽▼→↑↓↔◁◀▷▶♤♠♧♣⊙◈▣◐◑▒▤▥▨▧▦▩♨☏☎☞↕↗↙↖↘ª"
special_symbols_for_suffix = "←☜㏂㏘℡"
daggers = """†‡⸶⸷⸸⹋‖*∗"""
spaces = " \n\t\f\v\r\u200b\u200c\u2060\ufeff"

_url_prefix = "https|http|mailto|ftp"
_url_suffix = (
    "com|net|org|edu|game|ebiz|club|gov|mil|aero|asia|biz|cat|coop|info|int|jobs|mobi"
    "|museum|name|post|pro|tel|travel|xxx|ac|ad|ae|af|ag|ai|al|am|an|ao|aq|ar|as|at|au|aw|ax|az|ba|bb|bd"
    "|be|bf|bg|bh|bi|bj|bm|bn|bo|br|bs|bt|bv|bw|by|bz|ca|cc|cd|cf|cg|ch|ci|ck|cl|cm|cn|co|cr|cs|cu|cv|cx"
    "|cy|cz|dd|de|dj|dk|dm|do|dz|ec|ee|eg|eh|er|es|et|eu|fi|fj|fk|fm|fo|fr|ga|gb|gd|ge|gf|gg|gh|gi|gl|gm"
    "|gn|gp|gq|gr|gs|gt|gu|gw|gy|hk|hm|hn|hr|ht|hu|id|ie|il|im|in|io|iq|ir|is|it|je|jm|jo|jp|ke|kg|kh|ki"
    "|km|kn|kp|kr|kw|ky|kz|la|lb|lc|li|lk|lr|ls|lt|lu|lv|ly|ma|mc|md|me|mg|mh|mk|ml|mm|mn|mo|mp|mq|mr|ms"
    "|mt|mu|mv|mw|mx|my|mz|na|nc|ne|nf|ng|ni|nl|no|np|nr|nu|nz|om|pa|pe|pf|pg|ph|pk|pl|pm|pn|pr|ps|pt|pw"
    "|py|qa|re|ro|rs|ru|rw|sa|sb|sc|sd|se|sg|sh|si|sj|Ja|sk|sl|sm|sn|so|sr|ss|st|su|sv|sx|sy|sz|tc|td|tf"
    "|tg|th|tj|tk|tl|tm|tn|to|tp|tr|tt|tv|tw|tz|ua|ug|uk|us|uy|uz|va|vc|ve|vg|vi|vn|vu|wf|ws|ye|yt|yu|za"
    "|zm|zw|mp3|mp4|avi|html|idv|jpg|bmp|png|gif|htm|cdn|media"
)
url_pattern = re.compile(
    rf"""(?i)\b((?:{_url_prefix}?:(?:/{{1,3}}|[a-z0-9%])|[a-z0-9.\-]+[.](?:{_url_suffix})/)(?:[^\s()<>{{}}\[\]]+|\([^\s()]*?\([^\s()]+\)[^\s()]*?\)|\([^\s]+?\))+(?:\([^\s()]*?\([^\s()]+\)[^\s()]*?\)|\([^\s]+?\)|[^\s`!()\[\]{{}};:'".,<>?«»“”‘’])|(?:(?<!@)[a-z0-9]+(?:[.\-][a-z0-9]+)*[.](?:{_url_suffix})/?(?!@)))""",
)
email_pattern = re.compile(
    r"[a-z0-9.\-+_]+@[a-z0-9.\-+_]+\.[a-z]+|[a-z0-9.\-+_]+@[a-z0-9.\-+_]+\.[a-z]+\.[a-z]"
)

whitespaces = [" ", "\n", "\t", "\v", "\f", "\r"]
whitespaces_wo_space = ["\n", "\t", "\v", "\f", "\r"]

sf_exception = [
    "no",
    "No",
    "pp",
    "PP",
    "vol",
    "Vol",
    "vols",
    "Vols",
    "al",
    "ed",
    "Ed",
    "Eds",
    "trans",
    "Trans",
    "rev",
    "Rev",
    "p",
    "P",
    "n.p",
    "N.P",
    "N.p",
    "n.d",
    "N.D",
    "N.d",
    "page",
    "Page",
    "para",
    "Para",
    "comp",
    "Comp",
    "Capt",
    "capt",
    "dept",
    "Dept",
    "Mr",
    "mr",
    "Miss",
    "Mrs",
    "mrs",
    "Ms",
    "ms",
    "Dr",
    "dr",
    "Prof",
    "prof",
    "Rev",
    "rev",
    "St",
    "st",
    "Co",
    "co",
    "MM",
    "mm",
    "Messrs",
    "messrs",
    "Mlle",
    "mlle",
    "Mme",
    "mme",
    "def",
    "Def",
    "viz",
    "Viz",
]

for i in range(0, 10):
    sf_exception += [
        f"{i}항",
        f"{i}조",
        f"{i}호",
        f"{i}절",
        f"{i}목",
        f"{i}권",
        f"{i}쪽",
        f"{i}장",
        f"{i}",
    ]

sf_exception += [
    c
    for c in list(sorted(lower_alphabets))[:10]
    + list(sorted(upper_alphabets))[:10]
    + ["ㄱ", "ㄴ", "ㄷ", "ㄹ", "ㅁ", "ㅂ", "ㅅ", "ㅇ", "ㅈ", "ㅊ", "ㅋ", "ㅌ", "ㅍ", "ㅎ"]
]

sf_exception = [
    f" {w}{s.replace(' ', '')}" if " " in s else f"{w}{s}"
    for s in sf_exception
    for w in whitespaces
]


# apostrophe: I`m, You’re, ...
alphabet_with_quotes = {}
for alpha in alphabets:
    alphabet_with_quotes[alpha] = {
        f"{alpha}{q}{apo}" for apo in alphabets for q in ["'", "’", "`"]
    }

backup_normal = {
    ":)",
    ":(",
    ":'(",
    "O:)",
    "&)",
    ">:(",
    "3:)",
    '<(")',
    ":-)",
    ":-(",
    "관야유적",
    "라요 바예카노",
    "알림 차단",
    "알림 수신",
    "알림 문자",
    "에스타디오 비센테",
    "George D. Beauchamp",
    "Monkey D. Luffy",
    "몽키 D. 루피",
    "'N'",
    "'n'",
    "N' ",
    "n' ",
}

circle_bracket_charaters = [
    "⓪",
    "①",
    "②",
    "③",
    "④",
    "⑤",
    "⑥",
    "⑦",
    "⑧",
    "⑨",
    "⑩",
    "⑪",
    "⑫",
    "⑬",
    "⑭",
    "⑮",
    "⑯",
    "⑰",
    "⑱",
    "⑲",
    "⑳",
    "㉑",
    "㉒",
    "㉓",
    "㉔",
    "㉕",
    "㉖",
    "㉗",
    "㉘",
    "㉙",
    "㉚",
    "㉛",
    "㉜",
    "㉝",
    "㉞",
    "㉟",
    "㊱",
    "㊲",
    "㊳",
    "㊴",
    "㊵",
    "㊶",
    "㊷",
    "㊸",
    "㊹",
    "㊺",
    "㊻",
    "㊼",
    "㊽",
    "㊾",
    "㊿",
    "⓿",
    "❶",
    "❷",
    "❸",
    "❹",
    "❺",
    "❻",
    "❼",
    "❽",
    "❾",
    "❿",
    "⓫",
    "⓬",
    "⓭",
    "⓮",
    "⓯",
    "⓰",
    "⓱",
    "⓲",
    "⓳",
    "⓴",
    "⑴",
    "⑵",
    "⑶",
    "⑷",
    "⑸",
    "⑹",
    "⑺",
    "⑻",
    "⑼",
    "⑽",
    "⑾",
    "⑿",
    "⒀",
    "⒁",
    "⒂",
    "⒃",
    "⒄",
    "⒅",
    "⒆",
    "⒇",
    "⒜",
    "⒝",
    "⒞",
    "⒟",
    "⒠",
    "⒡",
    "⒢",
    "⒣",
    "⒤",
    "⒥",
    "⒦",
    "⒧",
    "⒨",
    "⒩",
    "⒪",
    "⒫",
    "⒬",
    "⒭",
    "⒮",
    "⒯",
    "⒰",
    "⒱",
    "⒲",
    "⒳",
    "⒴",
    "⒵",
    "Ⅰ",
    "Ⅱ",
    "Ⅲ",
    "Ⅳ",
    "Ⅴ",
    "Ⅵ",
    "Ⅶ",
    "Ⅷ",
    "Ⅸ",
    "Ⅹ",
    "Ⅺ",
    "Ⅻ",
    "ⅰ",
    "ⅱ",
    "ⅲ",
    "ⅳ",
    "ⅴ",
    "ⅵ",
    "ⅶ",
    "ⅷ",
    "ⅸ",
    "ⅹ",
    "ⅺ",
    "ⅻ",
    "☐",
    "☑",
    "☒",
    "Ⓐ",
    "Ⓑ",
    "Ⓒ",
    "Ⓓ",
    "Ⓔ",
    "Ⓕ",
    "Ⓖ",
    "Ⓗ",
    "Ⓘ",
    "Ⓩ",
    "Ⓚ",
    "Ⓛ",
    "Ⓜ",
    "Ⓝ",
    "Ⓞ",
    "Ⓟ",
    "Ⓠ",
    "Ⓡ",
    "Ⓢ",
    "Ⓣ",
    "Ⓤ",
    "Ⓥ",
    "Ⓦ",
    "Ⓧ",
    "Ⓨ",
    "ⓐ",
    "ⓑ",
    "ⓒ",
    "ⓓ",
    "ⓔ",
    "ⓕ",
    "ⓖ",
    "ⓗ",
    "ⓘ",
    "ⓙ",
    "ⓚ",
    "ⓛ",
    "ⓜ",
    "ⓝ",
    "ⓞ",
    "ⓟ",
    "ⓠ",
    "ⓡ",
    "ⓢ",
    "ⓣ",
    "ⓤ",
    "ⓥ",
    "ⓦ",
    "ⓧ",
    "ⓨ",
    "ⓩ",
    "🅐",
    "🅑",
    "🅒",
    "🅓",
    "🅔",
    "🅕",
    "🅖",
    "🅗",
    "🅘",
    "🅙",
    "🅚",
    "🅛",
    "🅜",
    "🅝",
    "🅞",
    "🅟",
    "🅠",
    "🅡",
    "🅢",
    "🅣",
    "🅤",
    "🅥",
    "🅦",
    "🅧",
    "🅨",
    "🅩",
    "㉠",
    "㉡",
    "㉢",
    "㉣",
    "㉤",
    "㉥",
    "㉦",
    "㉧",
    "㉨",
    "㉩",
    "㉪",
    "㉫",
    "㉬",
    "㉭",
    "㉮",
    "㉯",
    "㉰",
    "㉱",
    "㉲",
    "㉳",
    "㉴",
    "㉵",
    "㉶",
    "㉷",
    "㉸",
    "㉹",
    "㉺",
    "㉻",
    "⓵",
    "⓶",
    "⓷",
    "⓸",
    "⓹",
    "⓺",
    "⓻",
    "⓼",
    "⓽",
    "⓾",
]



# Copyright (C) 2021 Hyunwoong Ko <kevin.ko@tunib.ai> and Sang-Kil Park <skpark1224@hyundai.com>
# All rights reserved.

from kss._elements.element import Element


class Syllable(Element):
    def __init__(self, text, pos, idx):
        if len(text) != 1:
            raise ValueError("Length of syllable must be 1.")
        super().__init__(text, pos, idx)



# Copyright (C) 2021 Hyunwoong Ko <kevin.ko@tunib.ai> and Sang-Kil Park <skpark1224@hyundai.com>
# All rights reserved.

from typing import List

from kss._elements.subclasses import Syllable
from kss._modules.sentences.sentence_preprocessor import SentenceProcessor
from kss._utils.const import quotes_or_brackets_close_to_open, spaces, daggers


class SentencePostprocessor(SentenceProcessor):
    def postprocess(
        self,
        output_sentences: List[List[Syllable]],
        strip: bool,
    ) -> List[str]:
        """
        Postprocess output sentences by splitting rules

        Args:
            output_sentences (List[List[Syllable]]): output sentences by splitting rules in syllable object
            strip (bool): strip all sentences or not

        Returns:
            List[str]: postprocessed output setences in string
        """
        output_sentences = self._remove_first_space(output_sentences)
        output_sentences = self._remove_space_before_emoji(output_sentences)
        output_sentences = self._merge_broken_sub_sentence_in_quotes_or_brackets(
            output_sentences
        )
        output_sentences = self._move_first_footnote_in_sentence_to_previous(
            output_sentences
        )
        output_sentences = self._move_first_daggers_in_sentence_to_previous(
            output_sentences
        )
        output_sentences = self._move_non_structural_sub_sent_in_brackets_to_previous(
            output_sentences
        )
        output_sentences = self._move_unexpected_split_sentences_to_previous(
            output_sentences
        )
        output_sentences = self._move_symbol_sentences_only_to_previous(
            output_sentences
        )
        output_sentences = self._convert_syllables_to_sentences_with_cleaning(
            output_sentences, strip
        )
        return output_sentences

    def _merge_broken_sub_sentence_in_quotes_or_brackets(
        self, output_sentences: List[List[Syllable]]
    ) -> List[List[Syllable]]:
        """
        Merge broken sub-sentence in quotes or brackets.

        Args:
            output_sentences (List[List[Syllable]]): list of syllables

        Returns:
            List[List[Syllable]]: corrected list of syllables.

        Notes:
            열고 닫음이 분명한 괄호나 따옴표 안의 인용문/독백 등이 분리 된 경우 이를 찾아내서 병합한다.

            예시:
                입력: ["나는 생각했다. ", "(분명히 맞을 것이다. ", "그래야만 한다.) ", "그리곤 말했다."]
                출력: ["나는 생각했다. ", "(분명히 맞을 것이다. 그래야만 한다.) ", "그리곤 말했다."]
        """
        num_sentences = len(output_sentences)
        last_open, first_close = {}, {}

        for sentence_idx in range(num_sentences):
            found_close = False
            for syllable_idx, output_syllable in enumerate(
                output_sentences[sentence_idx]
            ):
                if output_syllable.check_pos("SSO", "QTO"):
                    if output_syllable.text not in last_open:
                        last_open[output_syllable.text] = [(sentence_idx, syllable_idx)]
                    else:
                        last_open[output_syllable.text].append(
                            (sentence_idx, syllable_idx)
                        )

                if output_syllable.check_pos("SSC", "QTC") and not found_close:
                    if output_syllable.text not in first_close:
                        first_close[output_syllable.text] = [
                            (sentence_idx, syllable_idx)
                        ]
                    else:
                        first_close[output_syllable.text].append(
                            (sentence_idx, syllable_idx)
                        )

            for syllable_close, close_values in first_close.items():
                if (
                    len(close_values) == 0
                    or syllable_close not in quotes_or_brackets_close_to_open
                ):
                    continue

                close_sent_idx, close_idx = close_values[-1]
                syllable_open = quotes_or_brackets_close_to_open[syllable_close]

                if syllable_open not in last_open:
                    continue

                open_values = [
                    (sent_i, syl_i)
                    for sent_i, syl_i in last_open[syllable_open]
                    if sent_i < close_sent_idx
                    or (sent_i == close_sent_idx and syl_i < close_idx)
                ]
                if len(open_values) == 0:
                    continue
                open_sent_idx, open_idx = open_values[-1]

                if close_sent_idx <= open_sent_idx:
                    continue
                for sub_sent_idx in range(open_sent_idx + 1, close_sent_idx + 1):
                    sub_sent = output_sentences[sub_sent_idx]
                    if sub_sent_idx == close_sent_idx:
                        sub_sent = sub_sent[: close_idx + 1]

                    for sub_syllable in sub_sent:
                        output_sentences[open_sent_idx].append(sub_syllable)

                    if sub_sent_idx == close_sent_idx:
                        output_sentences[sub_sent_idx] = output_sentences[sub_sent_idx][
                            close_idx + 1 :
                        ]
                    else:
                        output_sentences[sub_sent_idx] = []

                last_open[syllable_open].remove((open_sent_idx, open_idx))
                first_close[syllable_close].remove((close_sent_idx, close_idx))
        return self._remove_empty_sentence(output_sentences)

    @staticmethod
    def _check_text_from_character(output_syllable: Syllable, target: str):
        """
        Check given text is matched from character

        Args:
            output_syllable (Syllable): output syllable
            target (str): target string

        Returns:
            bool: match or not
        """
        _next = output_syllable
        if (not target.startswith(output_syllable.text)) and _next.text in target:
            split = target.split(_next.text)
            if len(split) > 1:
                target = output_syllable.text + target.split(_next.text)[1]

        for idx, char in enumerate(target):
            if _next.text != char:
                return False
            else:
                _next = _next.next
        return True

    def _move_first_daggers_in_sentence_to_previous(
        self, output_sentences: List[List[Syllable]]
    ) -> List[List[Syllable]]:
        """
        Move first daggers in sentence to previous.

        Args:
            output_sentences (List[List[Syllable]]): list of syllables

        Returns:
            List[List[Syllable]]: corrected list of syllables.

        Notes:
            칼표 처리:
                칼표가 처음으로 등장한 문장의 이전문장이 행갈이나 공백으로 끝난게 아니라면 칼표를 이전 문장으로 옮긴다.

            예시:
                입력: ["GPT3는 인공지능 모델이다.", "† 그러나 이 모델은""]
                출력: ["GPT3는 인공지능 모델이다.†", "그러나 이 모델은""] <--- 본문에 쓰인 각주로 인식.

                입력: ["GPT3는 인공지능 모델이다.\n", "† 그러나 이 모델은""]
                출력: ["GPT3는 인공지능 모델이다.\n", "† 그러나 이 모델은""]  <--- footer에 쓰인 각주로 인식. (그대로 유지)

                입력: ["GPT3는 인공지능 모델이다. ", "† 그러나 이 모델은""]
                출력: ["GPT3는 인공지능 모델이다. ", "† 그러나 이 모델은""]  <--- footer에 쓰인 각주로 인식. (그대로 유지)
        """

        for sentence_idx, output_sentence in enumerate(output_sentences):
            if sentence_idx != 0 and len(output_sentence) != 0:
                if output_sentence[0].next_skip_from_current("SP").text in daggers:
                    if output_sentences[sentence_idx - 1][-1].text not in " \r\n\v\f":
                        insert_idx = sentence_idx - 1
                        while insert_idx > 0 and len(output_sentences[insert_idx]) == 0:
                            insert_idx -= 1
                        output_sentences[insert_idx].append(output_sentence[0])
                        output_sentences[sentence_idx] = output_sentence[1:]

        return self._remove_empty_sentence(output_sentences)

    def _move_symbol_sentences_only_to_previous(
        self, output_sentences: List[List[Syllable]]
    ) -> List[List[Syllable]]:
        """
        Move symbol only sentences to previous

        Args:
            output_sentences (List[List[Syllable]]): list of syllables

        Returns:
            List[List[Syllable]]: corrected list of syllables.

        Notes:
            Symbol 처리:
                Symbol로만 이루어진 문장을 이전 문장으로 옮긴다.
        """
        for sentence_idx, output_sentence in enumerate(output_sentences):
            if (
                sentence_idx != 0
                and len(output_sentence) != 0
                and (
                    all(
                        [
                            syllable.check_pos("SY", "SF", "SE", "SC", "QT", "SS", "SP")
                            for syllable in output_sentence
                        ]
                    )
                )
            ):
                for output_syllable in output_sentence:
                    insert_idx = sentence_idx - 1
                    while insert_idx > 0 and len(output_sentences[insert_idx]) == 0:
                        insert_idx -= 1
                    output_sentences[insert_idx].append(output_syllable)
                output_sentences[sentence_idx] = []
        return self._remove_empty_sentence(output_sentences)

    def _move_first_footnote_in_sentence_to_previous(
        self, output_sentences: List[List[Syllable]]
    ) -> List[List[Syllable]]:
        """
        Move first footnote in sentence to previous.

        Args:
            output_sentences (List[List[Syllable]]): list of syllables

        Returns:
            List[List[Syllable]]: corrected list of syllables.

        Notes:
            각주 처리:
                최초로 발견되는 대괄호('[') 안에 있는 것이 각주이면 이를 이전 문장으로 옮긴다.

            예시:
                입력: ["그것은 사실이였다.", "[13] 하지만 그에 따라"]
                출력: ["그것은 사실이였다.[13]", "하지만 그에 따라"]

                출력: ["그것은 사실이였다.", "[편집]하지만 그에 따라"]
                출력: ["그것은 사실이였다.[편집]", "하지만 그에 따라"]

                출력: ["그것은 사실이였다.", "[편집]을 하고 싶지만 그에 따라"]
                출력: ["그것은 사실이였다.", "[편집]을 하고 싶지만 그에 따라"] <--- 유지

                출력: ["그것은 사실이였다.", "[더 보기] 마침 좋은 생각이 떠올랐다."]
                출력: ["그것은 사실이였다.[더 보기]", "마침 좋은 생각이 떠올랐다."]

                출력: ["그것은 사실이였다.", "[더 보기] 버튼을 눌러보세요."]
                출력: ["그것은 사실이였다.", "[더 보기] 버튼을 눌러보세요."] <--- 유지

                출력: ["그것은 사실이였다.", "[사각형]은"]
                출력: ["그것은 사실이였다.", "[사각형]은"] <--- 유지
        """

        for sentence_idx, output_sentence in enumerate(output_sentences):
            if sentence_idx != 0 and len(output_sentence) != 0:
                if output_sentence[0].next_skip_from_current("SP").text == "[":
                    close_idx = None
                    move = False
                    for syllable_idx, output_syllable in enumerate(output_sentence):
                        if output_sentences[sentence_idx - 1][
                            -1
                        ].text not in "\r\n\v\f" and (
                            output_syllable.text in "[0123456789*, ]"
                            or self._check_text_from_character(output_syllable, "편집]")
                            or self._check_text_from_character(output_syllable, "더 보기]")
                            or self._check_text_from_character(output_syllable, "더보기]")
                            or self._check_text_from_character(output_syllable, "스포일러]")
                            or self._check_text_from_character(output_syllable, "참고 ")
                        ):
                            move = True
                        else:
                            break

                        if output_syllable.text == "]":
                            close_idx = syllable_idx

                    if close_idx is not None:
                        if move is True:
                            if close_idx + 1 < len(output_sentence):
                                next_syllable = output_sentence[
                                    close_idx + 1
                                ].next_skip_from_current("SP")

                                move = (not next_syllable.pos.startswith("J")) and (
                                    not next_syllable.check_texts("버튼")
                                )

                        if move is True:
                            insert_idx = sentence_idx - 1
                            while (
                                insert_idx > 0
                                and len(output_sentences[insert_idx]) == 0
                            ):
                                insert_idx -= 1
                            output_sentences[insert_idx] += output_sentence[
                                : close_idx + 1
                            ]
                            output_sentences[sentence_idx] = output_sentence[
                                close_idx + 1 :
                            ]

        return self._remove_empty_sentence(output_sentences)

    def _move_non_structural_sub_sent_in_brackets_to_previous(
        self, output_sentences: List[List[Syllable]]
    ) -> List[List[Syllable]]:
        """
        Move non-structural sub sentence in brackets to previous.

        Args:
            output_sentences (List[List[Syllable]]): list of syllables

        Returns:
            List[List[Syllable]]: corrected list of syllables.

        Notes:
            최초로 발견되는 괄호('(') 안의 서브 문장이 명사(N*)로 끝나면서 공백을 제외한 다음 문자가
            조사(J*)가 아니라면 잘못 분리된 것으로 간주하고 이를 이전 문장으로 옮긴다.

            예시:
                입력: ["아니거든 !!! ", "(강한부정) 너가 먼자 말했잖아!"]
                출력: ["아니거든 !!! (강한부정)", " 너가 먼자 말했잖아!"]
        """
        for sentence_idx, output_sentence in enumerate(output_sentences):
            if sentence_idx != 0 and len(output_sentence) != 0:
                if output_sentence[0].next_skip_from_current("SP").text == "(":
                    close_idx, close_last = None, None
                    for syllable_idx, output_syllable in enumerate(output_sentence):
                        if output_syllable.text == ")":
                            close_idx = syllable_idx
                            close_last = syllable_idx == len(output_sentence) - 1
                            break

                    if close_idx is None:
                        continue

                    noun_finish = (
                        output_sentence[close_idx - 1]
                        .prev_skip_from_current(
                            *self._all_s_poses, exclude=self._all_s_exclude
                        )
                        .pos.startswith("N")
                    )

                    not_josa_start = close_last is True or not (
                        output_sentence[close_idx + 1]
                        .next_skip_from_current("SP")
                        .pos.startswith("J")
                    )

                    additional_merge = 0
                    if not close_last:
                        for output_syllable in output_sentence[close_idx + 1 :]:
                            if output_syllable.check_pos(
                                *self._all_s_poses, exclude=self._all_s_exclude
                            ):
                                additional_merge += 1
                            else:
                                break

                    if noun_finish and not_josa_start:
                        for output_syllable in output_sentence[
                            : close_idx + 1 + additional_merge
                        ]:
                            insert_idx = sentence_idx - 1
                            while (
                                insert_idx > 0
                                and len(output_sentences[insert_idx]) == 0
                            ):
                                insert_idx -= 1
                            output_sentences[insert_idx].append(output_syllable)
                        output_sentences[sentence_idx] = output_sentence[
                            close_idx + 1 + additional_merge :
                        ]

        return self._remove_empty_sentence(output_sentences)

    def _move_unexpected_split_sentences_to_previous(
        self, output_sentences: List[List[Syllable]]
    ) -> List[List[Syllable]]:
        """
        Move unexpected split sentences to previous.

        Args:
            output_sentences (List[List[Syllable]]): list of syllables

        Returns:
            List[List[Syllable]]: corrected list of syllables.

        Notes:
            분리된 문장이 조사(J*), 긍정지정사(VCP), 연결어미(EC), 보조용언(VX)으로 시작되면 이전 문장에 이어 붙인다.
            또는 문장이 칼표나 겹칼표 등의 주석 문자로 시작 했을때 다음 문자가 공백이 아니면

            예시:
                입력: ['반드시 막아야만 한다', '라고 했다']
                출력: ['반드시 막아야만 한다라고 했다']

                입력: ['반드시 막아야만 한다', '라고 했다']
                출력: ['반드시 막아야만 한다라고 했다']
        """
        for sentence_idx, output_sentence in enumerate(output_sentences):
            if (
                sentence_idx != 0
                and len(output_sentence) != 0
                and (
                    output_sentence[0].next_skip_from_current("SP").pos
                    in ("VCP+EC", "VX+EC")
                    or output_sentence[0]
                    .next_skip_from_current("SP")
                    .check_pos(
                        "J",
                        "VCP",
                        "EC",
                        "VX",
                        exclude=("MAJ", "+J", "+VCP", "+EC", "+VX", "JAMO"),
                    )
                )
            ):
                for output_syllable in output_sentence:
                    insert_idx = sentence_idx - 1
                    while insert_idx > 0 and len(output_sentences[insert_idx]) == 0:
                        insert_idx -= 1
                    output_sentences[insert_idx].append(output_syllable)
                output_sentences[sentence_idx] = []
        return self._remove_empty_sentence(output_sentences)

    @staticmethod
    def _convert_syllables_to_sentences_with_cleaning(
        output_sentences: List[List[Syllable]],
        strip: bool,
    ) -> List[str]:
        """
        Convert syllables to sentences with cleaning

        Args:
            output_sentences (List[List[Syllable]]): output syllables
            strip (bool): strip all sentences or not

        Returns:
            List[str]: output sentences list

        Notes:
            Syllable 객체를 모두 string으로 변경하고 각 문장에 strip을 수행한다.
        """
        final_output_sentences = []
        for output_sentence in output_sentences:
            output_sentence = "".join([syllable.text for syllable in output_sentence])
            if strip is True:
                output_sentence = output_sentence.strip(spaces)
            if len(output_sentence) != 0:
                final_output_sentences.append(output_sentence)

        return final_output_sentences

    @staticmethod
    def _remove_empty_sentence(output_sentences: List[List[Syllable]]):
        """
        Remove emtpy sentences after postprocessing

        Args:
            output_sentences (List[List[Syllable]]): split list of syllables

        Returns:
            List[List[Syllable]]: list of syllables without empty one
        """
        return [
            sentence
            for sentence in output_sentences
            if len("".join([i.text for i in sentence]).strip()) != 0
        ]

    @staticmethod
    def _remove_space_before_emoji(
        output_sentences: List[List[Syllable]],
    ) -> List[List[Syllable]]:
        """
        Remove a space character before emoji.
        The space character was appended in preprocessing step.

        Args:
            output_sentences (List[List[Syllable]]): split list of syllables

        Returns:
            List[List[Syllable]]: list of syllables without space before emoji
        """
        for sentence_idx, output_sentence in enumerate(output_sentences):
            for syllable_idx, output_syllable in enumerate(output_sentence):
                if (
                    output_syllable.check_pos("EMOJI")
                    and output_syllable.prev.text == " "
                ):
                    output_syllable.prev = output_syllable.prev.prev
                    output_sentences[sentence_idx].pop(syllable_idx - 1)
        return output_sentences

    @staticmethod
    def _remove_first_space(
        output_sentences: List[List[Syllable]],
    ) -> List[List[Syllable]]:
        """
        Remove a space added by preprocessor

        Args:
            output_sentences (List[List[Syllable]]): split list of syllables

        Returns:
            List[List[Syllable]]: list of syllables without added space
        """
        for sentence_idx, output_sentence in enumerate(output_sentences):
            for syllable_idx, output_syllable in enumerate(output_sentence):
                if (
                    sentence_idx == 0
                    and syllable_idx == 0
                    and output_syllable.check_pos("SP")
                ):
                    output_sentences[sentence_idx].pop(syllable_idx)

        return output_sentences



# This code was copied from Text Rank KR [https://github.com/theeluwin/textrankr]
# And modified by Hyunwoong Ko [https://github.com/hyunwoongko]

from collections import Counter
from itertools import combinations
from typing import List

from kss import split_morphemes, split_sentences
from kss._modules.summarization.sentence import Sentence


def _parse_text_into_sentences(
    text: str, backend: str, strip: bool, ignores: List[str]
) -> List[Sentence]:
    """
    This function splits the given text into sentence candidates using a pre-defined splitter,
    then creates a list of `sentence.Sentence` instances which have bag-of-words inside, tokenized by the given tokenizer.
    """

    # init
    index: int = 0
    duplication_checker: set = set()
    sentences: List[Sentence] = []

    # parse text
    candidates: List[str] = split_sentences(
        text, backend=backend, num_workers=1, strip=strip, ignores=ignores
    )
    for candidate in candidates:
        # cleanse the candidate
        if not len(candidate):
            continue
        if candidate in duplication_checker:
            continue

        # tokenize the candidate
        tokens: List[str] = split_morphemes(candidate, backend=backend, num_workers=1)
        if len(tokens) < 2:
            continue
        duplication_checker.add(candidate)

        # create a sentence
        bow: Counter = Counter(tokens)
        sentence = Sentence(index, candidate, bow)
        sentences.append(sentence)
        index += 1

    # return
    return sentences


def _multiset_jaccard_index(counter1: Counter, counter2: Counter) -> float:
    """
    Calculates the jaccard index between two given multisets.
    Note that a `Counter` instance can be used for representing multisets.
    """
    intersection_count: int = sum((counter1 & counter2).values())
    union_count: int = sum((counter1 | counter2).values())
    try:
        return intersection_count / union_count
    except ZeroDivisionError:
        return 0.0


def _build_sentence_graph(sentences: List[Sentence], tolerance: float):
    """
    Builds a `networkx.Graph` instance, using sentences as nodes.
    An edge weight is determined by the jaccard index between two sentences,
    but the edge will be ignored if the weight is lower than the given tolerance.
    """
    from networkx import Graph

    # init
    graph: Graph = Graph()

    # add nodes
    graph.add_nodes_from(sentences)

    # add edges
    for sentence1, sentence2 in combinations(sentences, 2):
        weight: float = _multiset_jaccard_index(sentence1.bow, sentence2.bow)
        if weight > tolerance:
            graph.add_edge(sentence1, sentence2, weight=weight)

    # return
    return graph



# Copyright (C) 2021 Hyunwoong Ko <kevin.ko@tunib.ai> and Sang-Kil Park <skpark1224@hyundai.com>
# All rights reserved.

from typing import List, Tuple

from kss._elements.subclasses import Syllable
from kss._modules.sentences.sentence_processor import SentenceProcessor
from kss._utils.const import (
    bracket_open_to_close,
    bracket_close_to_open,
    quotes_open_to_close,
    quotes_close_to_open,
    special_symbols_for_split,
    jamo,
    spaces,
    special_symbols_for_suffix,
    daggers,
    circle_bracket_charaters, whitespaces, whitespaces_wo_space,
)
from kss._utils.emojis import _emojis


class SentencePreprocessor(SentenceProcessor):

    _correction = {
        lambda c, p: c in circle_bracket_charaters: "SN",
        lambda c, p: c in special_symbols_for_split or c in daggers: "PF",  # Prefix
        lambda c, p: c in "!?.": "SF",
        lambda c, p: c in ",:/ㆍ": "SC",
        lambda c, p: c in "".join(bracket_open_to_close): "SSO",
        lambda c, p: c in "".join(bracket_close_to_open): "SSC",
        lambda c, p: c in "`'\"″": "QTN",  # quotes normal
        lambda c, p: c in "".join(quotes_open_to_close): "QTO",  # quotes open
        lambda c, p: c in "".join(quotes_close_to_open): "QTC",  # quotes close
        lambda c, p: c in jamo: "JAMO",
        lambda c, p: c == "요" and ("EC" in p or "JX" == p): "EF",
        lambda c, p: (
            c == "^" or c in _emojis or c in special_symbols_for_suffix
        ): "EMOJI",
        lambda c, p: c in spaces: "SP",
    }

    def preprocess(self, input_morphemes: List[Tuple[str, str]]) -> List[Syllable]:
        """
        Convert mecab morphemes to syllables and correct wrong tags

        Args:
            input_morphemes (List[Tuple[str, str]]): input morphemes

        Returns:
            List[Syllable]: syllables in input text
        """
        syllables = self._convert_morphemes_to_syllables(input_morphemes)
        syllables = self._correct_wrong_tags(syllables)
        syllables = self._append_space_before_emoji(syllables)
        return syllables

    def _convert_morphemes_to_syllables(
        self, input_morphemes: List[Tuple[str, str]]
    ) -> List[Syllable]:
        """
        Convert mecab morphemes to syllables.

        Args:
            input_morphemes (List[Tuple[str, str]]): input morphemes

        Returns:
            List[Syllable]: syllables in input text.
        """

        char_idx = 1
        prev = Syllable(" ", "SP", idx=0)
        syllables = [prev]

        for pos in input_morphemes:
            for char in pos[0]:
                tag = pos[1]
                for _func, _tag in self._correction.items():
                    if _func(char, pos[1]):
                        tag = _tag
                        break

                syllable = Syllable(char, tag, idx=char_idx)
                syllable.prev = prev
                prev.next = syllable
                syllables.append(syllable)

                prev = syllable
                char_idx += 1

        return syllables

    def _correct_wrong_tags(self, syllables: List[Syllable]):
        """
        Convert mecab morphemes to syllables and preprocess syllables

        Args:
            syllables (List[Syllable]): input syllables

        Returns:
            List[Syllable]: syllables in input text
        """
        for syllable in syllables:
            if syllable.check_pos_and_text(
                "JKS", "이"
            ) and syllable.next.check_pos_and_text("MAG", "다"):
                self._change_poses(syllable, "VCP", "EF")

            if syllable.check_pos_and_text(
                "EF", "네"
            ) and syllable.next.check_pos_and_text("XSN", "용"):
                self._change_poses(syllable, "EF", "EF")

            if syllable.check_pos_and_text(
                "EC", "까"
            ) and syllable.next.check_pos_and_text("NNG", "용"):
                self._change_poses(syllable, "EF", "EF")

            if (
                syllable.check_pos_and_text("EF", "을")
                and syllable.next.check_pos_and_text("EF", "까")
                and syllable.next.next.check_pos_and_text("XSN", "용")
            ):
                self._change_poses(syllable, "EF", "EF", "EF")

            if (
                syllable.check_pos_and_text("EP", "였")
                and syllable.next.check_pos_and_text("EC", "게")
                and syllable.next.next.check_pos_and_text("NNG", "용")
            ):
                self._change_poses(syllable, "EP", "EF", "EF")

            if syllable.check_pos_and_text(
                "EC", "구"
            ) and syllable.next.check_pos_and_text("NNG", "용"):
                self._change_poses(syllable, "EF", "EF")

            if syllable.check_pos_and_text(
                "EF", "엇"
            ) and syllable.next.check_pos_and_text("IC", "음"):
                self._change_poses(syllable, "EP", "ETN")

            if syllable.check_pos_and_text("EC", "쥬"):
                self._change_poses(syllable, "EF")

            if syllable.check_pos_and_text(
                "EC", "어"
            ) and syllable.next.check_pos_and_text("EC", "용"):
                self._change_poses(syllable, "EF", "EF")

            if syllable.check_pos_and_text("UNKNOWN", "떄"):
                self._change_poses(syllable, "NNG")

        return syllables

    @staticmethod
    def _change_poses(syllable: Syllable, *poses: str):
        """
        Change poses from the given syllable.
        This method could make a huge problem, so this implemented in preprocessor class.

        Args:
            syllable (Syllable): input syllable
            *poses (str): poses to be changed
        """
        _next = syllable
        for pos in poses:
            _next.pos = pos
            _next = _next.next

    @staticmethod
    def _append_space_before_emoji(syllables: List[Syllable]) -> List[Syllable]:
        """
        Append a space character before emoji character.
        This could be helpful for tokenizing sentences which contain emoji.

        Args:
            syllables (List[Syllable]): input syllables

        Returns:
            List[Syllable]: preprocessed syllables
        """
        new_syllables = []
        for syllable in syllables:
            if syllable.check_pos("EMOJI") and not syllable.prev.check_pos("EMOJI"):
                space_syllable = Syllable(" ", "SP", idx=syllable.idx)
                if syllable.prev is not None:
                    space_syllable.prev = syllable.prev
                space_syllable.next = syllable
                syllable.prev = space_syllable

                new_syllables.append(space_syllable)
            new_syllables.append(syllable)
        return new_syllables



# Copyright (C) 2021 Hyunwoong Ko <kevin.ko@tunib.ai> and Sang-Kil Park <skpark1224@hyundai.com>
# All rights reserved.

from functools import partial, lru_cache
from typing import List, Union, Tuple

from kss._elements.subclasses import Syllable
from kss._modules.morphemes.analyzers import Analyzer
from kss._modules.sentences.embracing_processor import EmbracingProcessor
from kss._modules.sentences.sentence_postprocessor import SentencePostprocessor
from kss._modules.sentences.sentence_preprocessor import SentencePreprocessor
from kss._modules.sentences.sentence_splitter import SentenceSplitter
from kss._utils.multiprocessing import _run_job
from kss._utils.sanity_checks import (
    _check_num_workers,
    _check_text,
    _check_analyzer_backend,
    _check_type,
    _check_iterable_type,
)

preprocessors = {(): SentencePreprocessor()}
postprocessors = {(): SentencePostprocessor()}


def split_sentences(
    text: Union[str, List[str], Tuple[str]],
    backend: str = "auto",
    num_workers: Union[int, str] = "auto",
    strip: bool = True,
    ignores: List[str] = None,
) -> Union[List[str], List[List[str]]]:
    """
    Split texts into sentences.

    Args:
        text (Union[str, List[str], Tuple[str]]): single text or list/tuple of texts
        backend (str): morpheme analyzer backend. 'mecab', 'pecab', 'punkt' are supported.
        num_workers (Union[int, str])): the number of multiprocessing workers
        strip (bool): strip all sentences or not
        ignores (List[str]): list of strings to ignore

    Returns:
        Union[List[str], List[List[str]]]: outputs of sentence splitting.
    """
    if ignores is None:
        ignores = []

    text, finish = _check_text(text)
    strip = _check_type(strip, "strip", bool)
    ignores = _check_iterable_type(ignores, "ignores", list, str)

    if finish:
        return text

    backend = _check_analyzer_backend(backend)
    num_workers = _check_num_workers(text, num_workers)

    ignores_tuple = tuple(ignores)
    if ignores_tuple not in preprocessors:
        preprocessors[ignores_tuple] = SentencePreprocessor(ignores)
        postprocessors[ignores_tuple] = SentencePostprocessor(ignores)
    _preprocessor = preprocessors[ignores_tuple]
    _postprocessor = postprocessors[ignores_tuple]

    return _run_job(
        func=partial(
            _split_sentences,
            backend=backend,
            strip=strip,
            preprocessor=_preprocessor,
            postprocessor=_postprocessor,
        ),
        inputs=text,
        num_workers=num_workers,
    )


@lru_cache(maxsize=500)
def _split_sentences(
    text: Union[str, Tuple[Syllable]],
    backend: Analyzer,
    strip: bool,
    postprocess: bool = True,
    recursion: int = 0,
    preprocessor: SentencePreprocessor = preprocessors[()],
    postprocessor: SentencePostprocessor = postprocessors[()],
) -> List[str]:
    """
    Split texts into sentences.

    Args:
        text (Union[str, List[Syllable]]): single text
        backend (str): morpheme analyzer backend
        strip (bool): strip all sentences or not
        postprocess (bool): whether it uses postprocessing or not
        recursion (int): recursion times

    Returns:
        List[str]: outputs of sentence splitting.
    """

    embracing = EmbracingProcessor()

    # 1. analyze morphemes
    if isinstance(text, str):
        backup_sentence = preprocessor.backup(text)
        morphemes = backend.pos(backup_sentence, drop_space=False)
        syllables = preprocessor.preprocess(morphemes)
    elif isinstance(text, tuple) and len(text) > 0 and isinstance(text[0], Syllable):
        syllables = text
    elif isinstance(text, tuple) and len(text) == 0:
        syllables = tuple()
    else:
        raise ValueError("Wrong data type input for `_split_sentences`.")

    # 2. define variables used for splitting
    output_sentences = []
    current_sentence_syllables = []
    prev_embracing_mode = False
    split_mode = False

    # 3. split sentences
    for idx, syllable in enumerate(syllables):
        sent_idx = len(output_sentences)
        splitter = SentenceSplitter(syllable)
        syllable_added = False
        embracing.process(idx, sent_idx, syllable)
        current_embracing_mode = not embracing.empty()

        if split_mode is False:
            if splitter.check_split_right_now():
                output_sentences.append(current_sentence_syllables)
                current_sentence_syllables = [syllable]
                syllable_added = True

            else:
                if backend._backend == "character":
                    split_mode = splitter._sf()
                else:
                    split_mode = splitter.check_split_start()

        else:
            end_split, end_split_exception = splitter.check_split_end()
            embracing.update_index(idx, sent_idx, syllable)

            if end_split is True:
                split_mode = False

                if current_embracing_mode is False and prev_embracing_mode is True:
                    current_sentence_syllables.append(syllable)
                    output_sentences.append(current_sentence_syllables)
                    current_sentence_syllables = []
                    syllable_added = True

                else:
                    if prev_embracing_mode is False:
                        output_sentences.append(current_sentence_syllables)
                        current_sentence_syllables = []
                    else:
                        split_mode = end_split_exception

        if not syllable_added:
            current_sentence_syllables.append(syllable)

        prev_embracing_mode = current_embracing_mode

    if len(current_sentence_syllables) != 0:
        output_sentences.append(current_sentence_syllables)

    # 4. realign wrong quotes and brackets
    if recursion < 10:
        output_sentences = embracing.realign(
            input_sentences=syllables,
            output_sentences=output_sentences,
            func=partial(
                _split_sentences,
                backend=backend,
                strip=strip,
                postprocess=False,
                recursion=recursion + 1,
            ),
        )

    # 5. postprocess
    if postprocess is True:
        output_sentences = postprocessor.postprocess(output_sentences, strip)
        output_sentences = [postprocessor.restore(s, text) for s in output_sentences]

    return output_sentences



# Copyright (C) 2021 Hyunwoong Ko <kevin.ko@tunib.ai> and Sang-Kil Park <skpark1224@hyundai.com>
# All rights reserved.

import platform
import unicodedata
from typing import Union, Tuple, List, Any, Optional, Type, Callable, Iterable

from kss._modules.morphemes.analyzers import (
    MecabAnalyzer,
    PecabAnalyzer,
    Analyzer,
    CharacterAnalyzer,
)
from kss._utils.logging import logger

MECAB_INFORM, KONLPY_MECAB_INFORM, PECAB_INFORM, PUNCT_INFORM = (
    False,
    False,
    False,
    False,
)

_mecab_info_linux_macos = "https://github.com/hyunwoongko/python-mecab-kor"
_konlpy_info_linux_macos = "https://konlpy.org/en/latest/api/konlpy.tag/#mecab-class"
_mecab_info_windows = "https://cleancode-ws.tistory.com/97"
_konlpy_info_windows = "https://uwgdqo.tistory.com/363"
_pecab_info = "https://github.com/hyunwoongko/pecab"


def _message_by_user_os(linux_macos: str, windows: str) -> str:
    user_os = platform.uname().system.lower()
    if user_os in ["linux", "darwin"]:
        return linux_macos
    elif user_os.startswith("win"):
        return windows


def _check_value(
    param: Any, param_name: str, predicate: Callable, suggestion: str
) -> Any:
    """
    Check param value

    Args:
        param (bool): param value
        param_name (str): param name
        predicate (Callable): callable
        suggestion (str): suggestion message

    Returns:
        Any: param value
    """
    available = predicate(param)

    if not available:
        raise TypeError(
            f"Oops! '{param}' is not supported value for `{param_name}`.\n"
            f"Currently kss only supports {suggestion} for this.\n"
            f"Please check `{param_name}` parameter again ;)\n"
        )

    return param


def _check_iterable_type(
    param: Iterable, param_name: str, iterable_type: Type, element_type: Type
) -> Any:
    """
    Check iterable param type

    Args:
        param (Iterable): iterable param value
        param_name (str): param name
        iterable_type (Type): iterable type
        element_type (Type): element type

    Returns:
        Any: param value
    """
    if not isinstance(param, iterable_type):
        raise TypeError(
            f"Oops! '{type(param)}' is not supported type for `{param_name}`.\n"
            f"Currently kss only supports {iterable_type} for this.\n"
            f"Please check `{param_name}` parameter again ;)\n"
        )

    for element in param:
        if not isinstance(element, element_type):
            raise TypeError(
                f"Oops! '{type(param).__qualname__.capitalize()}[{type(element).__qualname__}]' is not supported type for `{param_name}`.\n"
                f"Currently kss only supports {iterable_type.__qualname__.capitalize()}[{element_type.__qualname__}] for this.\n"
                f"Please check `{param_name}` parameter again ;)\n"
            )
    return param


def _check_type(param: Any, param_name: str, types: Union[Type, List[Type]]) -> Any:
    """
    Check param type

    Args:
        param (bool): param value
        param_name (str): param name
        types (Union[Type, List[Type]]): types

    Returns:
        Any: param value
    """
    was_list = True

    if not isinstance(types, list) or not isinstance(types, tuple):
        types = [types]
        was_list = False

    available = any([isinstance(param, _type) for _type in types])

    if not available:
        raise TypeError(
            f"Oops! '{type(param)}' is not supported type for `{param_name}`.\n"
            f"Currently kss only supports {types if was_list else types[0]} {'types' if was_list else 'type'} for this.\n"
            f"Please check `{param_name}` parameter again ;)\n"
        )

    return param


def _check_text(
    text: Union[str, List[str], Tuple[str]]
) -> Tuple[Union[str, List[str], Tuple[str]], bool]:
    """
    Check input text type.

    Args:
        text (Union[str, List[str], Tuple[str]]): single text or list/tuple of texts

    Returns:
        Tuple[Union[str, List[str], Tuple[str]], bool]: single text or list/tuple of texts
            and whether it can finish processing right now or not.
    """
    finish = False

    if (
        not isinstance(text, str)
        and not isinstance(text, list)
        and not isinstance(text, tuple)
    ):
        raise TypeError(
            f"Oops! '{type(text)}' is not supported type for `text`.\n"
            f"Currently kss only supports [str, List[str], Tuple[str]] types for this.\n"
            f"Please check `text` parameter again ;)\n"
        )

    elif (isinstance(text, list) or isinstance(text, tuple)) and len(text) != 0:
        are_strings = [isinstance(t, str) for t in text]
        if not all(are_strings):
            not_string_idx = are_strings.index(False)
            raise TypeError(
                "Oops! Some elements in `text` were not string.\n"
                f"For example, index {not_string_idx} was {type(text[not_string_idx])}.\n"
                f"Currently kss only supports [str, List[str], Tuple[str]] for this.\n"
                f"Please check `text` parameter again ;)\n"
            )
        else:
            text = [unicodedata.normalize("NFC", t) for t in text]

        # unwrap list or tuple with single text
        if len(text) == 1:
            text = text[0]

        if isinstance(text, list):
            text = tuple(text)

    elif not isinstance(text, str) and len(text) == 0:
        finish = True

    elif isinstance(text, str):
        text = unicodedata.normalize("NFC", text)

    return text, finish


def _check_analyzer_backend(backend: str) -> Analyzer:
    """
    Check morpheme analyzer backend type.

    Args:
        backend (str):

    Returns:
        Analyzer: morpheme analyzer backend.
    """
    global MECAB_INFORM, KONLPY_MECAB_INFORM, PECAB_INFORM, PUNCT_INFORM

    if isinstance(backend, str):
        backend = backend.lower()

    if backend not in ["auto", "mecab", "pecab", "punct"]:
        raise ValueError(
            f"Oops! '{backend}' is not supported value for `backend`.\n"
            f"Currently kss only supports ['auto', 'pecab', 'mecab', 'punct'] for this.\n"
            f"Please check `backend` parameter again ;)\n"
        )

    mecab_backend = MecabAnalyzer()
    pecab_backend = PecabAnalyzer()
    punct_backend = CharacterAnalyzer()

    if backend == "punct":
        return punct_backend

    if backend == "mecab":
        if mecab_backend._backend is not None:
            return mecab_backend
        else:
            raise ImportError(
                _message_by_user_os(
                    linux_macos="Oops! You specified `backend` as 'mecab', but you don't have mecab in your environment.\n"
                    "If you want to use mecab backend, please install mecab or konlpy.tag.Mecab!\n"
                    "Please refer to following web sites for details:\n"
                    f"- mecab: {_mecab_info_linux_macos}\n"
                    f"- konlpy.tag.Mecab: {_konlpy_info_linux_macos}\n",
                    windows="Oops! You specified `backend` as 'mecab', but you don't have mecab in your environment.\n"
                    "If you want to use mecab backend, please install mecab or konlpy.tag.Mecab!\n"
                    "Please refer to following web sites for details:\n"
                    f"- mecab: {_mecab_info_windows}\n"
                    f"- konlpy.tag.Mecab: {_konlpy_info_windows}\n",
                )
            )

    elif backend == "pecab":
        if pecab_backend._backend is not None:
            return pecab_backend
        else:
            raise ImportError(
                "Oops! You specified `backend` as 'pecab', but you don't have pecab in your environment.\n"
                "If you want to use pecab backend, please install pecab!\n"
                "Please refer to following web sites for details:\n"
                "- pecab: https://github.com/hyunwoongko/pecab\n"
            )

    if backend == "auto":
        if mecab_backend._backend == "mecab":
            if not MECAB_INFORM:
                logger.warning(
                    "Oh! You have mecab in your environment. Kss will take this as a backend! :D\n"
                )
                MECAB_INFORM = True
            return mecab_backend

        elif mecab_backend._backend == "konlpy":
            if not KONLPY_MECAB_INFORM:
                logger.warning(
                    "Oh! You have konlpy.tag.Mecab in your environment. Kss will take this as a backend! :D\n"
                )
                KONLPY_MECAB_INFORM = True
            return mecab_backend

        elif pecab_backend._backend == "pecab":
            if not PECAB_INFORM:

                installation_help_message = _message_by_user_os(
                    linux_macos="For your information, Kss also supports mecab backend.\n"
                    "We recommend you to install mecab or konlpy.tag.Mecab for faster execution of Kss.\n"
                    "Please refer to following web sites for details:\n"
                    f"- mecab: {_mecab_info_linux_macos}\n"
                    f"- konlpy.tag.Mecab: {_konlpy_info_linux_macos}\n",
                    windows="For your information, Kss also supports mecab backend.\n"
                    "We recommend you to install mecab or konlpy.tag.Mecab for faster execution of Kss.\n"
                    "Please refer to following web sites for details:\n"
                    f"- mecab: {_mecab_info_windows}\n"
                    f"- konlpy.tag.Mecab: {_konlpy_info_windows}\n",
                )

                logger.warning(
                    "Because there's no supported C++ morpheme analyzer, "
                    "Kss will take pecab as a backend. :D\n"
                    f"{installation_help_message}"
                )
                PECAB_INFORM = True

            return pecab_backend
        else:
            if not PUNCT_INFORM:

                installation_help_message = _message_by_user_os(
                    linux_macos="For your information, Kss also supports mecab backend.\n"
                    "We recommend you to install mecab or konlpy.tag.Mecab for faster execution of Kss.\n"
                    "Please refer to following web sites for details:\n"
                    f"- mecab: {_mecab_info_linux_macos}\n"
                    f"- konlpy.tag.Mecab: {_konlpy_info_linux_macos}\n",
                    windows="For your information, Kss also supports mecab backend.\n"
                    "We recommend you to install mecab or konlpy.tag.Mecab for faster execution of Kss.\n"
                    "Please refer to following web sites for details:\n"
                    f"- mecab: {_mecab_info_windows}\n"
                    f"- konlpy.tag.Mecab: {_konlpy_info_windows}\n",
                )

                logger.warning(
                    "Because there's no supported C++ morpheme analyzer, "
                    "Kss will take pecab as a backend. :D\n"
                    f"{installation_help_message}"
                )
                PUNCT_INFORM = True

            return punct_backend


def _check_num_workers(
    inputs: Any, num_workers: Union[int, str]
) -> Optional[Union[int, bool]]:
    """
    Check the number of multiprocessing workers.

    Args:
        inputs (Any): input data
        num_workers (Union[int, str]): the number of multiprocessing workers

    Returns:
        Optional[Union[int, bool]]: the number of multiprocessing workers.
            `None` means maximum number of workers which can be used for now.
            `False` means that it will not use multiprocessing.
    """

    if isinstance(num_workers, float):
        num_workers = int(num_workers)
    elif isinstance(num_workers, str):
        num_workers = num_workers.lower()

    if not isinstance(num_workers, int) and num_workers != "auto":
        raise TypeError(
            f"Oops! '{num_workers}' is not supported value for `num_workers`.\n"
            f"Currently kss only supports [int, 'auto'] for this.\n"
            f"Please check `num_workers` parameter again ;)"
        )
    elif isinstance(num_workers, int) and num_workers < 1:
        raise ValueError(
            f"Oops! `num_workers` must be same or greater than 1, but you input {num_workers}.\n"
            "- You can input 1 for no multiprocessing.\n"
            "- You can input 2 ~ n for multiprocessing.\n"
            "- You can input 'auto' for using maximum workers.\n"
            "Please check `num_workers` parameter again ;)"
        )

    if (
        isinstance(inputs, str)
        or (
            (isinstance(inputs, list) or isinstance(inputs, tuple)) and len(inputs) == 1
        )
        or num_workers == 1
    ):
        return False
        # no multiprocessing worker

    elif num_workers == "auto":
        num_workers = None
        # maximum multiprocessing workers

    return num_workers



# Copyright (C) 2021 Hyunwoong Ko <kevin.ko@tunib.ai> and Sang-Kil Park <skpark1224@hyundai.com>
# All rights reserved.

from functools import partial
from typing import List, Union, Tuple

from kss._utils.multiprocessing import _run_job
from kss._utils.sanity_checks import (
    _check_text,
    _check_analyzer_backend,
    _check_num_workers,
    _check_type,
)


def split_morphemes(
    text: Union[str, List[str], Tuple[str]],
    backend: str = "auto",
    num_workers: Union[int, str] = "auto",
    drop_space: bool = True,
    return_pos: bool = True,
) -> Union[List[Tuple[str, str]], List[List[Tuple[str, str]]], Union[List, Tuple]]:
    """
    Split texts into morphemes.

    Args:
        text (Union[str, List[str], Tuple[str]]): single text or list/tuple of texts
        backend (str): morpheme analyzer backend. 'mecab', 'pecab' are supported.
        num_workers (Union[int, str])): the number of multiprocessing workers
        drop_space (bool): drop all spaces in output or not

    Returns:
        Union[List[Tuple[str, str]], List[List[Tuple[str, str]]], Union[List, Tuple]]:
            outputs of morpheme analysis.
    """
    text, finish = _check_text(text)
    drop_space = _check_type(drop_space, "drop_space", bool)

    if finish:
        return text

    num_workers = _check_num_workers(text, num_workers)
    backend = _check_analyzer_backend(backend)
    result = _run_job(partial(backend.pos, drop_space=drop_space), text, num_workers)
    if not return_pos:
        if isinstance(text, str):
            result = [token[0] for token in result]
        else:
            result = [[token[0] for token in tokens] for tokens in result]
    return result



# Copyright (C) 2021 Hyunwoong Ko <kevin.ko@tunib.ai> and Sang-Kil Park <skpark1224@hyundai.com>
# All rights reserved.

import multiprocessing as mp
from typing import Union, Any, List, Optional, Callable


def _run_job(
    func: Callable,
    inputs: Any,
    num_workers: Optional[Union[int, bool]] = None,
) -> Union[Any, List[Any]]:
    """
    Run job with or without multiprocessing.

    Args:
        func (Callable): function to run
        inputs (Any): input data
        num_workers (Optional[Union[int, bool]]): the number of multiprocessing workers.

    Returns:
        Union[Any, List[Any]]: output of the job.
    """
    if num_workers is False:
        if isinstance(inputs, str):
            output = func(inputs)
        else:
            output = [func(i) for i in inputs]
        return output
    else:
        with mp.Pool(num_workers) as pool:
            output = pool.map(func, inputs)
            pool.close()
            pool.join()
            return output



import codecs
import os
import subprocess
import sys
from contextlib import suppress

from setuptools import setup, find_packages
from setuptools.command.install import install


class InstallCommand(install):
    def run(self):
        try:
            import mecab
        except:
            with suppress():
                subprocess.call(
                    [sys.executable, "-m", "pip", "install", "python-mecab-kor"],
                    stderr=subprocess.DEVNULL,
                )

        super().run()


def read_file(filename, cb):
    with codecs.open(filename, "r", "utf8") as f:
        return cb(f)


version = None
with open(os.path.join("kss", "__init__.py"), encoding="utf-8") as f:
    for line in f:
        if line.strip().startswith("__version__"):
            version = line.split("=")[1].strip().replace('"', "").replace("'", "")

with open("README.md", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="kss",
    version=version,
    author="Hyunwoong Ko",
    author_email="kevin.ko@tunib.ai",
    url="https://github.com/hyunwoongko/kss",
    license='BSD 3-Clause "New" or "Revised" License',
    description="A Toolkit for Korean sentence segmentation",
    long_description_content_type="text/markdown",
    platforms=["any"],
    install_requires=["emoji==1.2.0", "regex", "pecab", "networkx"],
    long_description=long_description,
    packages=find_packages(exclude=["bench", "assets", ".java", ".pytest_cache"]),
    python_requires=">=3",
    zip_safe=False,
    package_data={"": ["kss/pynori/resources/*"]},
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    cmdclass={"install": InstallCommand},
)



# Copyright (C) 2021 Hyunwoong Ko <kevin.ko@tunib.ai> and Sang-Kil Park <skpark1224@hyundai.com>
# All rights reserved.

from functools import lru_cache
from typing import Tuple

from kss._elements.subclasses import Syllable
from kss._modules.sentences.sentence_processor import SentenceProcessor
from kss._utils.const import sf_exception, jaum


class SentenceSplitter(SentenceProcessor):
    """
    Sentence Splitting Rules class

    Args:
        syllable (Syllable): current syllable
    """

    unavailable_next = set()
    unavailable_next.update({"시피", "거나", "의 "})
    unavailable_next.update({"에" + add for add in ["서", "선", "는", "도 "]})
    unavailable_next.update(
        {"라" + add for add in ["서", "는", "도 ", "던", "지만", "고", "건", "거나", "며", "면서"]}
    )
    unavailable_next.update(
        {"하" + add for add in ["여서", "여도", "였", "고는 ", "고서 ", "곤 ", "다는", "는데.", "며"]}
    )
    unavailable_next.update({"할" + add for add in ["텐데"]})

    def __init__(self, syllable: Syllable):
        super().__init__()
        self.syllable = syllable

    def __hash__(self):
        """Hash function for lru_cache"""
        return hash(self.syllable)

    ####################
    # Flow Controllers #
    ####################

    def check_split_start(self) -> bool:
        """
        Check whether the given syllable is split point or not.

        Returns:
            bool: whether the given syllable is split point or not.

        Notes:
            분할 시작 규칙:
                종결부호(SF), 캐릭터 휴리스틱 규칙, 4개의 어말어미(EF, EC, ETN, ETM) 규칙 중 하나라도 성립하면 분할한다.
        """
        return (
            self._sf()
            or self._ef()
            or self._ec()
            or self._etn()
            or self._etm()
            or self._char()
        )

    def check_split_end(self) -> Tuple[bool, bool]:
        """
        Check whether the given syllable is split end point or not.

        Returns:
            bool: whether the given syllable is split end point or not.

        Notes:
            분할 종료 규칙:
                1. 현재 문자 이후의 문자가 자모(JAMO)이면 분할을 뒤로 미룬다.
                2. 현재 문자 이후의 문자가 대부분의 부호에 속하거나 현재 문자가 쉼표(,)인데 다음 문자도 쉼표(,)이면 분할을 뒤로 미룬다.
                3. 현재 문자 이후의 문자도 분할이 가능하면 분할을 뒤로 미룬다.
                4. 위 세가지 규칙에 해당 사항이 없는 경우 분할한다.

            예외:
                1. 만약 현재 문자가 물음표(?) 혹은 느낌표(!)인데 공백을 제외한 다음 문자가 구두점(.)이면 즉시 분할한다.
                2. 만약 현재 문자가 하이푼(-)인데 공백을 제외한 다음 문자가 종결부호(SF)가 아닌 경우 지금 즉시 분할한다.
                3. 만약 현재 문자가 자음인데 이전문자가 공백이고 다음 문자가 종결부호(SF)이며 다다음 문자가 공백이면 즉시 분할한다.
        """

        no_more_all_s = not self._check_pos(
            self._all_s_poses_wo_qtn, exclude=self._all_s_exclude
        )

        end_split = no_more_all_s and not (
            (self._check_text(",") and self._check_next_text(","))
            or (self._check_text(",") and self._check_prev_text(","))
        )

        end_split = end_split and not self.syllable.check_pos("JAMO")
        end_split = end_split and not self.check_split_start()
        end_split_exception = False

        if not end_split:
            # 예외 1
            if self._check_prev_skip_sp_text(
                ("?", "!")
            ) and self._check_next_skip_sp_text("."):
                end_split = True
                end_split_exception = True

            # 예외 2
            elif self._check_text(("－", "-", "–")) and not (
                self._check_next_skip_sp_pos("SF")
                or self._check_next_text(("－", "-", "–"))
            ):
                end_split = True
                end_split_exception = True

            elif (
                self.syllable.prev.check_pos("SP")
                and self.syllable.check_text(*jaum)
                and self.syllable.next.check_text(".")
                and self.syllable.next.next.check_text(" ")
            ):
                end_split = True
                end_split_exception = True

        return end_split, end_split_exception

    ###################
    # Splitting Rules #
    ###################

    def check_split_right_now(self) -> bool:
        """
        Check whether the given syllable is split end point or not.

        Returns:
            bool: whether the given syllable is split end point or not.

        Notes:
            단락기호(¶)가 등장하면 곧바로 분리한다.
        """
        available = self.syllable.check_text("¶")
        return available

    def _sf(self) -> bool:
        """
        Check whether the given syllable is split end point or not.

        Returns:
            bool: whether the given syllable is split end point or not.

        Notes:
            종결부호 분할 규칙:
                종결부호(SF) 뒤에 공백(SP) 혹은 기타 부호(SY), 열린 괄호(SSO), 이모지(EMOJI), 한글 자모(JAMO)가 존재하면 분할한다.

            예외:
                1. 현재 문자 앞의 공백(SP)과 종결부호(SF)을 제외한 이전 문자가 숫자이면 분할하지 않는다.
                2. 현재 문자 앞의 공백(SP)과 종결부호(SF)을 제외한 문자가 긍정지정사(VCP)이면 분할하지 않는다.
                3. 현재 문자 앞의 공백(SP)과 종결부호(SF)을 제외한 문자가 조사(J*)이면서 다음 문자가 구두점(.)이면 분할하지 않는다.
                4. 현재 문자 뒤의 공백(SP)과 종결부호(SF)을 제외한 문자가 긍정지정사(VCP)이면 분할하지 않는다.
                5. 현재 문자 앞의 공백(SP)과 종결부호(SF)을 제외한 문자가 접속부사(MAJ)이면 분할하지 않는다.
                6. 현재 문자 앞의 공백(SP)과 종결부호(SF)을 제외한 문자가 '만', '데+..' 등 이면 분할하지 않는다.
                7. 현재 문자 뒤의 공백(SP)을 제외한 문자가 마침표(.)가 아니고 그 뒤 문자가 곧 바로 마침표라면 분할하지 않는다.
                8. 현재 문자 뒤로 등장하는 한글 문자열이 몇가지 분할하지 않아야 하는 경우에 속하면 분할하지 않는다.
                9. 현재 문자가 ' no.', ' No.', ' vol.', ' p.', ' pp.', ' page.', ' al.', ' ed.', ' eds.'
                    ' 항.', ' 조.', ' 호.', ' 절.', ' 권.', " 쪽.' 등에 존재하면 분할하지 않는다.
        """

        available = False

        # 종결부호 분할 규칙
        if self._check_pos("SF") and (
            self._check_next_pos("SP")
            or self._check_next_skip_sp_pos(("SY", "SSO", "EMOJI", "JAMO"))
        ):
            # 예외 1
            available = not self._prev_skip(("SP", "SF")).text.isnumeric()

            # 예외 2
            available = available and not (
                self._check_prev_skip_spsf_pos("VCP", exclude="+VCP")
            )

            # 예외 3
            available = available and not (
                self._check_prev_skip_spsf_pos("J", exclude=("EMOJI", "JAMO", "+J"))
                and self._check_text(".")
                and self._check_prev_text(".")
            )

            # 예외 4
            available = available and not (
                self._check_next_skip_spsf_pos("VCP", exclude="+VCP")
            )

            # 예외 5
            available = available and not self._check_prev_skip_spsf_pos(("MAJ", "MAG"))

            # 예외 6
            available = available and not (
                (
                    self._check_prev_skip_spsf_pos("EC")
                    and self._check_prev_skip_spsf_text("만")
                )
                or (
                    self._check_prev_skip_spsf_pos(("EC", "NNB"))
                    and self._check_prev_skip_spsf_text("데")
                    and self._check_prev_text(".")
                )
            )

            # 예외 7
            next_skip_sp = self._next_skip("SP")
            prev_skip_sp = self._prev_skip("SP")

            available = (
                available
                and not (
                    not next_skip_sp.check_text(".")
                    and next_skip_sp.next_skip("SP").check_text(".")
                )
                and not (
                    not prev_skip_sp.check_text(".")
                    and prev_skip_sp.prev_skip("SP").check_text(".")
                )
            )

            # 예외 8
            available = available and not self._check_next_is_unavailable_split()

            # 예외 9
            available = available and not (
                self._check_text(".")
                and self._check_multiple_prev_texts_from_before(*sf_exception)
            )

        return available

    def _ef(self) -> bool:
        """
        Check whether the given syllable is split end point or not.

        Returns:
            bool: whether the given syllable is split end point or not.

        Notes:
            종결어미 분할 규칙:
                종결어미(EF)인 형태소의 마지막 문자이면 분할한다. 단, 현재 형태소에 조사(J*) 성분이 포함되면 분할하지 않는다.

            에외:
                1. 현재 문자 뒤의 공백(SP) 및 대부분 부호를 제외한 다음 문자가 쉼표(,)이면 분할하지 않는다. 단 ,,과 같이 연속되어 쓰이면 분할을 허용한다.
                2. 현재 문자 뒤의 공백(SP)을 제외한 다음 문자가 연결어미(EC), 조사(J*), 보조용언(VX) 중 하나이면 분할하지 않는다.
                3. 현재 형태소가 대명사(NP)+긍정지정사(VCP)+EF(종결어미)인데, 바로 뒤 문자가 종결부호(SF)가 아니면 분할하지 않는다.
                4. 현재 문자 바로 이전 문자가 공백(SP)이고 그 이전 문자가 보조용언(VX)이면 분할하지 않는다.
                5. 현재 문자 뒤로 등장하는 한글 문자열이 몇가지 분할하지 않아야 하는 경우에 속하면 분할하지 않는다.
        """

        available = False

        # 종결어미 분할 규칙
        if self._check_pos("EF", exclude="+J") and not self._check_next_skip_sp_pos(
            "EF"
        ):
            # 예외 1
            available = not self._check_non_doubled_comma()

            # 예외 2
            available = available and not (
                self._check_next_skip_sp_pos(
                    ("EC", "J", "VX"),
                    exclude=("MAJ", "EMOJI", "JAMO", "+VX", "+EC", "+J", "VX+"),
                )
            )

            # 예외 3
            if available and self._check_prev_pos("NP+VCP+EF"):
                available = self._check_next_pos("SF")

            # 예외 4
            available = available and not (
                self._check_prev_text(" ")
                and self.syllable.prev.check_pos("VX", exclude="VX+")
            )

            # 예외 5
            available = available and not self._check_next_is_unavailable_split()

        return available

    def _ec(self) -> bool:
        """
        Check whether the given syllable is split end point or not.

        Returns:
            bool: whether the given syllable is split end point or not.

        Notes:
            연결어미 분할 규칙:
                연결어미(EC)인 형태소의 마지막 문자가 '다'일 때, 다음 허용항을 만족하면 분할한다.
                본래, 연결어미(EC)는 분할하면 안되지만, 형태소 분석기의 성능 문제로 종결어미(EF)를 연결어미(EC)로 인식되는 경우에 대응하기 위함이다.
                단, 현재 형태소에 조사(J*) 성분이 포함되면 분할하지 않는다.

            허용:
                1. 현재 문자의 이전 문자가 선어말어미(EP)이고 현재 문자의 뒤의 공백(SP)을 제외한 다음 문자가 닫는 괄호(SSO), 닫는 따옴표(QTO), 부호(SF, SY, EMOJI) 중 하나이면 분할한다.
                2. 현재 문자의 이전 문자가 선어말어미(EP)이고 현재 문자 뒤의 공백(SP)을 제외한 다음 문자가 대명사(NP) 혹은 접속부사(MAJ)이면 분할한다.
                3. 현재 문자의 이전 문자가 연결어미(EC)이면서 '니'인데, 뒤의 문자가 '만'이면서 연결어미(EC)가 아니면 분할한다. 이는 '입니다' 및 '습니다'가 연결어미로 인식되는 경우에 대응하기 위함이다.

            예외:
                1. 현재 문자 뒤의 공백(SP) 및 대부분 부호를 제외한 다음 문자가 쉼표(,)이면 분할하지 않는다. 단 ,,과 같이 연속되어 쓰이면 분할을 허용한다.
                2. 현재 문자 뒤의 공백(SP)을 제외한 다음 문자가 긍정지정사(VCP) 조사(J*), 보조용언(VX) 중 하나이면 분할하지 않는다.
                3. 현재 문자 뒤로 등장하는 한글 문자열이 몇가지 분할하지 않아야 하는 경우에 속하면 분할하지 않는다.
        """
        available = False
        permission = False

        # 연결어미 분할 규칙
        if (
            self._check_text("다")
            and self._check_pos("EC", exclude="+J")
            and not self._check_next_skip_sp_pos("EC")
        ):
            # 예외 1
            available = not self._check_non_doubled_comma()

            # 예외 2
            available = available and not self._check_next_skip_sp_pos(
                ("VCP", "J", "VX"),
                exclude=("MAJ", "JAMO", "EMOJI", "+J", "+VCP", "+VX"),
            )

            # 예외 3
            available = available and not self._check_next_is_unavailable_split()

        if available:
            # 허용 1:
            permission = self._check_prev_pos("EP") and self._check_next_skip_sp_pos(
                ("SF", "SY", "SSO", "QTO", "EMOJI")
            )

            # 허용 2:
            permission = permission or (
                self._check_prev_pos("EP") and self._check_next_skip_sp_pos("NP", "MAJ")
            )

            # 허용 3:
            permission = permission or (
                self._check_prev_text("니")
                and self._check_prev_pos("EC")
                and not (self._check_next_text("만") and self._check_next_pos("EC"))
            )

        return available and permission

    def _etn(self) -> bool:
        """
        Check whether the given syllable is split end point or not.

        Returns:
            bool: whether the given syllable is split end point or not.

        Notes:
            명사형 전성어미 분할 규칙:
                 명사형 전성어미(ETN)인 형태소의 마지막 문자가 '기'나 '길'이 아니면 분할한다.
                 이는 명사형 전성어미가 종결형으로 사용 될 경우에 대응 하기 위함이다. (~함, ~음, ~됨 등)
                 단, 현재 형태소에 조사(J*) 및 접미사(XS*) 성분이 포함되면 분할하지 않는다.

            예외:
                1. 현재 문자 뒤의 공백(SP) 및 대부분 부호를 제외한 다음 문자가 쉼표(,)이면 분할하지 않는다. 단 ,,과 같이 연속되어 쓰이면 분할을 허용한다.
                2. 현재 문자 뒤의 바로 뒤 문자가 공백, 한글 자모(JAMO), 이모지, 부호 등이 아니면 분할하지 않는다.
                3. 현재 문자 뒤의 공백(SP)을 제외한 다음 문자가 조사(J*), 동사(VV), 형용사(VA), 보조용언(VA) 이면 분할하지 않는다.
                4. 현재 문자 뒤의 공백(SP)과 종결부호(SF)를 제외한 다음 문자가 관형사(MM) 혹은 닫는 괄호(SSC)이면 분할하지 않는다.
                5. 현재 문자 뒤의 명사형 전성어미(ETN)을 제외한 이전의 문자와 동일한 문자가 현재 문자의 바로 뒤에 등장하면 분할하지 않는다.
                6. 현재 문자 뒤로 등장하는 한글 문자열이 몇가지 분할하지 않아야 하는 경우에 속하면 분할하지 않는다.
        """
        available = False

        if (
            self._check_pos("ETN", exclude=("+J", "XS"))
            and not self._check_next_skip_sp_pos("ETN")
            and not self._check_text("기", "길")
        ):
            # 예외 1
            available = not self._check_non_doubled_comma()

            # 예외 2
            available = available and (
                self._check_next_pos(("SP", "SF", "SY", "SSO", "QTO", "EMOJI", "JAMO"))
            )

            # 예외 3
            available = available and not self._check_next_skip_sp_pos(
                ("J", "VV", "VA", "VX"),
                exclude=("MAJ", "EMOJI", "JAMO", "+J"),
            )

            # 예외 4
            available = available and not self._check_next_skip_spsf_pos(("MM", "SSC"))

            # 예외 5
            available = available and not (
                self._check_next_skip_all_s_text(
                    self._prev_skip("ETN")
                    .prev_skip(*self._all_s_poses, exclude=self._all_s_exclude)
                    .text
                )
            )

            # 예외 6
            available = available and not self._check_next_is_unavailable_split()

        return available

    def _etm(self) -> bool:
        """
        Check whether the given syllable is split end point or not.

        Returns:
            bool: whether the given syllable is split end point or not.

        Notes:
            관형형 전성어미 분할 규칙:
                관형형 전성어미(ETM)인 형태소의 마지막 문자이면 다음 허용항을 만족하면 분할한다.
                이는 관형형 전성어미가 종결형으로 사용 될 경우에 대응 하기 위함이다. (~다능, ~했다던... 등)
                단, 현재 형태소에 조사(J*) 및 접미사(XS*) 성분이 포함되면 분할하지 않는다.

            허용:
                1. 이전 문자가 '다' 이면서 현재 문자가 '능'이면 분할한다.
                2. 현재 문자 뒤의 공백(SP)을 제외한 다음 문자가 말줄임표(SE)이면 분할한다.

            예외:
                1. 현재 문자의 뒤의 공백(SP) 및 대부분 부호를 제외한 다음 문자가 쉼표(,)이면 분할하지 않는다. 단 ,,과 같이 연속되어 쓰이면 분할을 허용한다.
                2. 현재 문자 뒤의 공백(SP) 및 대부분의 부호를 제외한 다음 문자가 의존명사(NNB)이면 분할하지 않는다.
                3. 현재 문자 뒤의 공백(SP)을 제외한 다음 문자가 조사(J*)이면 분할하지 않는다.
                4. 현재 문자 뒤로 등장하는 한글 문자열이 몇가지 분할하지 않아야 하는 경우에 속하면 분할하지 않는다.
        """
        available = False
        permission = False

        if self._check_pos(
            "ETM", exclude=("+J", "XS")
        ) and not self._check_next_skip_sp_pos("ETM"):
            # 예외 1
            available = not self._check_non_doubled_comma()

            # 예외 2
            available = available and not self._check_next_skip_all_s_pos("NNB")

            # 예외 3
            available = available and not self._check_next_skip_sp_pos(
                "J", exclude=("MAJ", "EMOJI", "JAMO", "+J")
            )

            # 예외 4
            available = available and not self._check_next_is_unavailable_split()

        if available:
            # 허용 1
            permission = self._check_prev_texts("다능")

            # 허용 2
            permission = permission or self._check_next_skip_sp_pos("SE")

        return available and permission

    def _char(self):
        """
        Check whether the given syllable is split end point or not.

        Returns:
            bool: whether the given syllable is split end point or not.

        Notes:
            캐릭터 기반의 휴리스틱 분할 규칙:
                캐릭터 기반의 문장분리 휴리스틱 알고리즘 적용한다.

            허용:
                듯 규칙 1. 현재 문자가 의존명사(NNB)이면서 '듯'이고 공백(SP) 종결부호(SF)를 제외한 다음 문자가 한글 자모(JAMO), 이모지(EMOJI), 말 줄임표(SE)
                        중 하나이면 현재 문자 뒤로 이어지는 모든 자모(JAMO) 및 이모지(EMOJI) 뒤의 문자가 동사가 아닌 경우 분할한다.
        """

        # 듯 규칙 1.
        if (
            self._check_text("듯")
            and self._check_pos("NNB")
            and (
                self._check_next_skip_spsf_pos(("EMOJI", "JAMO"))
                or self._check_next_skip_spsf_text("…")
            )
        ):
            _next = self.syllable.next_skip("SP")
            _cur_cnt, _max_cnt = 0, 5
            while _cur_cnt < _max_cnt and not _next.check_pos("JAMO", "EMOJI"):
                _next = _next.next_skip("SP")
                _cur_cnt += 1
            if not _next.check_pos("VV"):
                return True

        return False

    #####################
    # Utility functions #
    #####################

    @staticmethod
    def _tuple(exclude):
        if isinstance(exclude, str):
            exclude = (exclude,)
        return exclude

    @lru_cache(1)
    def _prev(self):
        return self.syllable.prev

    @lru_cache(1)
    def _next(self):
        return self.syllable.next

    @lru_cache(30)
    def _prev_skip(
        self,
        poses=None,
        exclude=None,
    ):
        return self.syllable.prev_skip(
            *self._tuple(poses),
            exclude=self._tuple(exclude),
        )

    @lru_cache(30)
    def _next_skip(
        self,
        poses=None,
        exclude=None,
    ):
        return self.syllable.next_skip(
            *self._tuple(poses),
            exclude=self._tuple(exclude),
        )

    @lru_cache(30)
    def _check_pos(
        self,
        poses=None,
        exclude=None,
    ):
        return self.syllable.check_pos(
            *self._tuple(poses),
            exclude=self._tuple(exclude),
        )

    @lru_cache(30)
    def _check_text(
        self,
        texts=None,
        exclude=None,
    ):
        return self.syllable.check_text(
            *self._tuple(texts),
            exclude=self._tuple(exclude),
        )

    @lru_cache(30)
    def _check_next_pos(
        self,
        poses=None,
        exclude=None,
    ):
        return self.syllable.next.check_pos(
            *self._tuple(poses),
            exclude=self._tuple(exclude),
        )

    @lru_cache(30)
    def _check_prev_pos(
        self,
        poses=None,
        exclude=None,
    ):
        return self.syllable.prev.check_pos(
            *self._tuple(poses),
            exclude=self._tuple(exclude),
        )

    @lru_cache(30)
    def _check_next_text(
        self,
        texts=None,
        exclude=None,
    ):
        return self.syllable.next.check_text(
            *self._tuple(texts),
            exclude=self._tuple(exclude),
        )

    @lru_cache(30)
    def _check_prev_text(
        self,
        texts=None,
        exclude=None,
    ):
        return self.syllable.prev.check_text(
            *self._tuple(texts),
            exclude=self._tuple(exclude),
        )

    @lru_cache(30)
    def _check_next_skip_pos(
        self,
        poses=None,
        exclude=None,
        skip=None,
        skip_exclude=None,
    ):
        return self.syllable.next_skip(
            *self._tuple(skip),
            exclude=self._tuple(skip_exclude),
        ).check_pos(
            *self._tuple(poses),
            exclude=self._tuple(exclude),
        )

    @lru_cache(30)
    def _check_prev_skip_pos(
        self,
        poses=None,
        exclude=None,
        skip=None,
        skip_exclude=None,
    ):
        return self.syllable.prev_skip(
            *self._tuple(skip),
            exclude=self._tuple(skip_exclude),
        ).check_pos(
            *self._tuple(poses),
            exclude=self._tuple(exclude),
        )

    @lru_cache(30)
    def _check_next_skip_text(
        self,
        texts=None,
        exclude=None,
        skip=None,
        skip_exclude=None,
    ):
        return self.syllable.next_skip(
            *self._tuple(skip),
            exclude=self._tuple(skip_exclude),
        ).check_text(
            *self._tuple(texts),
            exclude=self._tuple(exclude),
        )

    @lru_cache(30)
    def _check_prev_skip_text(
        self,
        texts=None,
        exclude=None,
        skip=None,
        skip_exclude=None,
    ):
        return self.syllable.prev_skip(
            *self._tuple(skip),
            exclude=self._tuple(skip_exclude),
        ).check_text(
            *self._tuple(texts),
            exclude=self._tuple(exclude),
        )

    @lru_cache(30)
    def _check_next_skip_sp_pos(
        self,
        poses=None,
        exclude=None,
    ):
        return self.syllable.next_skip("SP").check_pos(
            *self._tuple(poses),
            exclude=self._tuple(exclude),
        )

    @lru_cache(30)
    def _check_next_skip_from_current_pos(
        self,
        poses=None,
        exclude=None,
    ):
        return self.syllable.next_skip_from_current("SP").check_pos(
            *self._tuple(poses),
            exclude=self._tuple(exclude),
        )

    @lru_cache(30)
    def _check_prev_skip_sp_pos(
        self,
        poses=None,
        exclude=None,
    ):
        return self.syllable.prev_skip("SP").check_pos(
            *self._tuple(poses),
            exclude=self._tuple(exclude),
        )

    @lru_cache(30)
    def _check_next_skip_spsf_pos(
        self,
        poses=None,
        exclude=None,
    ):
        return self.syllable.next_skip("SP", "SF").check_pos(
            *self._tuple(poses),
            exclude=self._tuple(exclude),
        )

    @lru_cache(30)
    def _check_prev_skip_spsf_pos(
        self,
        poses=None,
        exclude=None,
    ):
        return self.syllable.prev_skip("SP", "SF").check_pos(
            *self._tuple(poses),
            exclude=self._tuple(exclude),
        )

    @lru_cache(30)
    def _check_next_skip_sp_text(
        self,
        texts=None,
        exclude=None,
    ):
        return self.syllable.next_skip("SP").check_text(
            *self._tuple(texts),
            exclude=self._tuple(exclude),
        )

    @lru_cache(30)
    def _check_prev_skip_sp_text(
        self,
        texts=None,
        exclude=None,
    ):
        return self.syllable.prev_skip("SP").check_text(
            *self._tuple(texts),
            exclude=self._tuple(exclude),
        )

    @lru_cache(30)
    def _check_next_skip_spsf_text(
        self,
        texts=None,
        exclude=None,
    ):
        return self.syllable.next_skip("SP", "SF").check_text(
            *self._tuple(texts),
            exclude=self._tuple(exclude),
        )

    @lru_cache(30)
    def _check_prev_skip_spsf_text(
        self,
        texts=None,
        exclude=None,
    ):
        return self.syllable.prev_skip("SP", "SF").check_text(
            *self._tuple(texts),
            exclude=self._tuple(exclude),
        )

    @lru_cache(30)
    def _check_next_skip_all_s_pos(
        self,
        poses=None,
        exclude=None,
    ):
        return self.syllable.next_skip(
            *self._all_s_poses,
            exclude=self._all_s_exclude,
        ).check_pos(
            *self._tuple(poses),
            exclude=self._tuple(exclude),
        )

    @lru_cache(30)
    def _check_next_skip_all_s_text(
        self,
        texts=None,
        exclude=None,
    ):
        return self.syllable.next_skip(
            *self._all_s_poses,
            exclude=self._all_s_exclude,
        ).check_text(
            *self._tuple(texts),
            exclude=self._tuple(exclude),
        )

    @lru_cache(30)
    def _check_texts(self, texts=None):
        return self.syllable.check_texts(*self._tuple(texts))

    @lru_cache(30)
    def _check_prev_texts(self, texts=None):
        return self.syllable.prev.check_texts(*self._tuple(texts))

    @lru_cache(30)
    def _check_next_skip_all_s_texts(self, texts=None):
        return self.syllable.next_skip_from_current(
            *self._all_s_poses,
            exclude=self._all_s_exclude,
        ).check_texts(*self._tuple(texts))

    @lru_cache(30)
    def _check_next_skip_all_s_multiple_texts(self, *texts):
        next_skip_all_s = self.syllable.next_skip(
            *self._all_s_poses,
            exclude=self._all_s_exclude,
        )

        for text in texts:
            if next_skip_all_s.check_texts(*self._tuple(text)):
                return True

        return False

    @lru_cache(1)
    def _check_next_is_unavailable_split(self):
        return self._check_next_skip_all_s_multiple_texts(*self.unavailable_next)

    @lru_cache(1)
    def _check_non_doubled_comma(self):
        next_all_s = self._next_skip(self._all_s_poses, exclude=self._all_s_exclude)
        return next_all_s.check_text(",") and not next_all_s.next_skip("SP").check_text(
            ","
        )

    @lru_cache(30)
    def _check_prev_texts_from_before(self, text):
        _prev = self.syllable
        for _ in text:
            _prev = _prev.prev

        return _prev.check_texts(text)

    @lru_cache(30)
    def _check_multiple_prev_texts_from_before(self, *texts):
        return any(self._check_prev_texts_from_before(t) for t in texts)

    @lru_cache(30)
    def _check_multiple_next_texts_from_current(self, *texts):
        return any(self._check_texts(t) for t in texts)



# Copyright (C) 2021 Hyunwoong Ko <kevin.ko@tunib.ai> and Sang-Kil Park <skpark1224@hyundai.com>
# All rights reserved.

from dataclasses import dataclass
from functools import lru_cache
from typing import Optional, Tuple, Union

from kss._elements.empty import Empty


@dataclass
class Element(object):
    _next: "Element" = None
    _prev: "Element" = None

    def __init__(self, text, pos, idx):
        self.text = text
        self.pos = pos
        self.idx = idx

    def __str__(self):
        return str((self.text, self.pos, self.idx))

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return other.text == self.text and other.pos == self.pos and other.idx == self.idx

    def __hash__(self):
        return hash((self.text, self.pos, self.idx))

    @property
    def next(self):
        if self._next is None:
            return Empty()
        return self._next

    @next.setter
    def next(self, _next):
        self._next = _next

    @property
    def prev(self):
        if self._prev is None:
            return Empty()
        return self._prev

    @prev.setter
    def prev(self, _prev):
        self._prev = _prev

    def next_skip(self, *poses, exclude=None):
        _next = self.next

        if isinstance(_next, Empty):
            return _next

        while _next.check_pos(*poses, exclude=exclude):
            _next = _next.next

        return _next

    def prev_skip(self, *poses, exclude=None):
        _prev = self.prev

        if isinstance(_prev, Empty):
            return _prev

        while _prev.check_pos(*poses, exclude=exclude):
            _prev = _prev.prev

        return _prev

    def prev_skip_from_current(self, *poses, exclude=None):
        _prev = self

        if isinstance(_prev, Empty):
            return _prev

        while _prev.check_pos(*poses, exclude=exclude):
            _prev = _prev.prev

        return _prev

    def next_skip_from_current(self, *poses, exclude=None):
        _next = self

        if isinstance(_next, Empty):
            return _next

        while _next.check_pos(*poses, exclude=exclude):
            _next = _next.next

        return _next

    def check_pos(self, *poses, exclude: Optional[Tuple] = None) -> bool:
        """
        Check pos of given syllable.

        Args:
            poses (str): poses for check
            exclude (Optional[Tuple]): excluded poses

        Returns:
            bool: whether pos of the syllable is contained in input poses or not.
        """
        for target in poses:
            if target in self.pos:
                if exclude is not None:
                    for e in exclude:
                        if e in self.pos:
                            return False
                return True
        return False

    def check_text(self, *texts, exclude: Optional[Tuple] = None) -> bool:
        """
        Check text of given syllable.

        Args:
            texts (str): texts for check
            exclude (Optional[Tuple]): excluded texts

        Returns:
            bool: whether text of the syllable is contained in input texts or not.
        """
        for target in texts:
            if target in self.text:
                if exclude is not None:
                    for e in exclude:
                        if e in self.text:
                            return False
                return True
        return False

    def check_pos_and_text(
        self,
        poses: Union[str, Tuple],
        texts: Union[str, Tuple],
        exclude_poses: Optional[Tuple] = None,
        exclude_texts: Optional[Tuple] = None,
    ):
        """
        Check pos and text at the same time

        Args:
            poses: poses for check
            texts (Tuple): texts for check
            exclude_poses (Optional[Tuple]): excluded poses
            exclude_texts (Optional[Tuple]): excluded texts

        Returns:
            bool: whether pos and text of the syllable are contained in input poses and texts.
        """
        if isinstance(poses, str):
            poses = (poses,)
        if isinstance(texts, str):
            texts = (texts,)

        return self.check_pos(*poses, exclude=exclude_poses) and self.check_text(
            *texts, exclude=exclude_texts
        )

    def check_texts(self, text: str) -> bool:
        """
        Check texts of current and next syllables

        Args:
            text (str): texts for check

        Returns:
            bool: whether text of the current and next syllables re contained in input text or not.
        """
        _node = self
        matches = []
        for char in text:
            same = _node.text == char
            if same is False:
                return False
            else:
                matches.append(same)
                _node = _node.next
        return all(matches)



# Copyright (C) 2021 Hyunwoong Ko <kevin.ko@tunib.ai> and Sang-Kil Park <skpark1224@hyundai.com>
# All rights reserved.

from typing import List, Tuple


def _get_linux_mecab():
    from mecab import MeCab

    analyzer = MeCab()
    analyzer.morphs("_")
    backend = "mecab"
    return analyzer, backend


def _get_linux_konlpy_mecab():
    from konlpy.tag import Mecab

    analyzer = Mecab()
    analyzer.morphs("_")
    backend = "konlpy"
    return analyzer, backend


def _get_windows_mecab():
    import MeCab

    analyzer = MeCab.Tagger()

    def pos(text: str) -> List[Tuple[str, str]]:
        """Mecab MSVC wrapper"""
        node = analyzer.parseToNode(text)
        outputs = []

        while node:
            surface = node.surface
            tag = node.feature.split(",")[0]
            if "BOS" not in tag and "EOS" not in tag:
                outputs.append((surface, tag))
            node = node.next

        return outputs

    setattr(analyzer, "pos", pos)
    analyzer.pos("_")
    backend = "mecab"
    return analyzer, backend


def _get_windows_konlpy_mecab():
    from konlpy.tag import Mecab

    analyzer = Mecab(dicpath=r"C:\mecab\mecab-ko-dic")
    analyzer.pos("_")
    backend = "konlpy"
    return analyzer, backend


def _get_mecab():
    """
    Try to import and get mecab morpheme analyzer

    Returns:
        Tuple[Optional[Union[mecab.MeCab, konlpy.tag.Mecab, Mecab.Tagger]], Optional[str]]:
            mecab morpheme analyzer and its backend
    """
    try:
        # python-mecab-kor (Linux/MacOS)
        return _get_linux_mecab()
    except Exception:
        try:
            # konlpy (Linux/MacOS)
            return _get_linux_konlpy_mecab()
        except Exception:
            try:
                # mecab-python-msvc (Windows)
                return _get_windows_mecab()
            except Exception:
                try:
                    # konlpy (Windows)
                    return _get_windows_konlpy_mecab()
                except Exception:
                    return None, None


def _get_pecab():
    """
    Try to import and get pecab morpheme analyzer

    Returns:
        Tuple[Optional[pecab.PeCab], Optional[str]]:
            pecab morpheme analyzer and its backend
    """
    try:
        from pecab import PeCab

        analyzer = PeCab()
        analyzer.morphs("_")
        backend = "pecab"
        return analyzer, backend
    except Exception:
        return None, None


def _preserve_space(
    text: str,
    tokens: List[Tuple[str, str]],
    spaces: str,
) -> List[Tuple[str, str]]:
    """
    Restore spaces from analyzed results

    Args:
        text (str): input text
        tokens (List[Tuple[str, str]]): analyzed results
        spaces (str): space tokens to add

    Returns:
        List[Tuple[str, str]]: analyzed results with space
    """
    results = list()
    len_text = len(text)
    len_tokens = len(tokens)
    text_idx = 0
    token_idx = 0

    while text_idx < len_text:
        character = text[text_idx]
        if character in spaces:
            results.append((character, "SP"))
            text_idx += 1
        else:
            if token_idx < len_tokens:
                token = tokens[token_idx]
                results.append(token)
                text_idx += len(token[0])
                token_idx += 1
            else:
                break
    return results



# Copyright (C) 2021 Hyunwoong Ko <kevin.ko@tunib.ai> and Sang-Kil Park <skpark1224@hyundai.com>
# All rights reserved.

import logging
from logging import getLogger

logging.basicConfig(format="[Kss]: %(message)s", level=logging.WARNING)
logger = getLogger()
